# Copyright (C) 2020 SuperDARN Canada, University of Saskatchewan
# Author: Daniel Billett, Marina Schmidt
#
# Modifications:
# 2021-05-07: CJM - Included radar position and labels in plotting
# 2021-04-01 Shane Coyle added pcolormesh to the code
# 2021-05-19 Marina Schmidt - Added scan index with datetimes
# 2021-09-15 Marina Schmidt - removed fov file options
# 2021-09-09: CJM - Included a channel option for plot_fan
# 2021-09-08: CJM - Included individual gate and beam boundary plotting for FOV
# 2021-11-22: MTS - pass in axes object to plot_fov
# 2021-11-18: MTS - Added Projectsion class for cartopy use
# 2021-02-02: CJM - Included rsep and frang options in plot_fov
#                 - Re-arranged logic for ranges option
#                 - Indexing fixed for give lower ranges that are non-zero
# 2022-03-10: MTS - switched coords involving range estimations to
#                   RangeEstimation
#                 - Removed GEO_COASTALINE, replaced with GEO on projections
#                 - reduced some parameters inputs to kwargs
# 2022-03-22: CJM - Set cmap bad values to transparent
# 2022-03-23: MTS - added the NotImplementedError for AACGM and GEO projection
#                   as this has yet to be figured out
# 2023-02-06: CJM - Added option to plot single beams in a scan or FOV diagram
# 2023-03-01: CJM - Added ball and stick plotting options
# 2023-03-01: CJM - Added ball and stick plotting options (merged later in year)
# 2023-08-16: CJM - Corrected for winding order in geo plots
# 2023-06-28: CJM - Refactored return values
# 2023-10-14: CJM - Add embargoed data method
# 2024-10-09: DDB - Control marker and its size in plot_radar_position()
#
# Disclaimer:
# pyDARN is under the LGPL v3 license found in the root directory LICENSE.md
# Everyone is permitted to copy and distribute verbatim copies of this license
# document, but changing it is not allowed.
#
# This version of the GNU Lesser General Public License incorporates the terms
# and conditions of version 3 of the GNU General Public License,
# supplemented by the additional permissions listed below.


"""
Fan plots, mapped to AACGM coordinates in a polar format
"""

import datetime as dt
import matplotlib.pyplot as plt
import numpy as np
import warnings

from matplotlib import ticker, cm, colors, axes
from typing import List

# Third party libraries
import aacgmv2

from pydarn import (PyDARNColormaps, partial_record_warning,
                    time2datetime, plot_exceptions, SuperDARNRadars, RadarID,
                    calculate_azimuth, Projs, Coords,
                    find_records_by_datetime, find_records_by_scan,
                    determine_embargo, add_embargo)


class Fan:
    """
    'Fan', or 'Field-of-view' plots for SuperDARN FITACF data
    This class inherits from matplotlib to generate the figures
    Methods
    -------
    plot_fan
    plot_fov
    plot_radar_position
    plot_radar_label
    """

    def __str__(self):
        return "This class is static class that provides"\
                " the following methods: \n"\
                "   - plot_fan()\n"\
                "   - plot_fov()\n"

    @staticmethod
    def plot_fan(dmap_data: List[dict], ax=None, ranges=None,
                 scan_index: int = 1,
                 scan_time: dt.datetime = None,
                 scan_time_tolerance: dt.timedelta = dt.timedelta(seconds=30),
                 parameter: str = 'v', cmap: str = None,
                 groundscatter: bool = False, zmin: int = None,
                 zmax: int = None, colorbar: bool = True,
                 colorbar_label: str = '', cax=None,
                 title: bool = True, boundary: bool = True,
                 projs: Projs = Projs.POLAR,
                 coords: Coords = Coords.AACGM_MLT,
                 channel: int = 'all', ball_and_stick: bool = False,
                 len_factor: float = 300, beam: int = None, **kwargs):
        """
        Plots a radar's Field Of View (FOV) fan plot for the given data and
        scan number

        Parameters
        -----------
            dmap_data: List[dict]
                Named list of dictionaries obtained from SDarn_read
            ax: axes.Axes
                Pre-defined axis object to pass in, must currently be
                polar projection
                Default: Generates a polar projection for the user
                with MLT/latitude labels
            ranges: List[int]
                Range bounds to plot, as [lower_bound, upper_bound].
                Default: Plots all ranges out to max given in hardware file.
            scan_index: int
                Scan number starting from the first record in file with
                associated channel number
                Default: 1
            scan_time: dt.datetime
                Datetime of the scan you would like to plot. Overrides
                specification of scan_index.
            scan_time_tolerance: dt.timedelta
                Search radius when filtering scan by datetime. All records
                with a timestamp of scan_time +/- scan_time_tolerance will
                be plotted.
                Default: 30 seconds
            parameter: str
                Key name indicating which parameter to plot.
                Default: v (Velocity). Alternatives: 'p_l', 'w_l', 'elv'
            cmap: matplotlib.cm
                matplotlib colour map
                https://matplotlib.org/tutorials/colors/colormaps.html
                Default: Official pyDARN colour map for given parameter
            groundscatter : bool
                Set true to indicate if groundscatter should be plotted in grey
                Default: False
            zmin: int
                The minimum parameter value for coloring
                Default: {'p_l': [0], 'v': [-200], 'w_l': [0], 'elv': [0]}
            zmax: int
                The maximum parameter value for  coloring
                Default: {'p_l': [50], 'v': [200], 'w_l': [250], 'elv': [50]}
            colorbar: bool
                Draw a colourbar if True
                Default: True
            colorbar_label: str
                the label that appears next to the colour bar.
                Requires colorbar to be true
                Default: ''
            cax: axes.Axes
                Pre-defined axis for the colorbar. If not specified and colorbar
                is True, a new axis will be created.
                Default: None
            title: bool
                if true then will create a title, else user
                can define it with plt.title
                default: true
            boundary: bool
                if true then plots the FOV boundaries
                default: true
            projs: Enum
                choice of projection for plot
                default: Projs.POLAR (polar projection)
            coords: Enum
                choice of plotting coordinates
                default: Coords.AACGM_MLT (Magnetic Lat and MLT)
            beam : int or None
                integer indicating if the user would like to plot a single beam
            channel : int or str
                integer indicating which channel to plot or 'all' to
                plot all channels
                Default: 'all'
            ball_and_stick : bool
                plot the data as a vector instead of filling a box
                Default: False
            len_factor : float
                control the length of the ball and stick plot stick length
                Default : 300
            kwargs: key = value
                Additional keyword arguments to be used in projection plotting
                and plot_fov for possible keywords, see: projections.axis_polar

        Returns
        -----------
        beam_corners_aacgm_lats
            n_beams x n_gates numpy array of latitudes
            return values dependent on given coords enum
        beam_corners_aacgm_lons
            n_beams x n_gates numpy array of longitudes or MLT
            return values dependent on given coords enum
        scan
            n_beams x n_gates numpy array of the scan data
            (for the selected parameter)
        grndsct
            n_beams x n_gates numpy array of the scan data
            (for the selected parameter)
        See Also
        --------
            plot_fov
        """
        # Remove all data from dmap_data that is not in chosen channel
        if ranges is None:
            ranges = []
        if channel != 'all':
            # Get the first channel used in case of no data in given channel
            opt_channel = dmap_data[0]['channel']
            dmap_data = [rec for rec in dmap_data if rec['channel'] == channel]
            # If no records exist, advise user that the channel is not used
            if not dmap_data:
                raise plot_exceptions.NoChannelError(channel, opt_channel)
        # Get the records which match scan_index
        if scan_time is not None:
            matching_records = find_records_by_datetime(dmap_data,
                                                        scan_time,
                                                        scan_time_tolerance)
            date = scan_time
        else:
            matching_records = find_records_by_scan(dmap_data, scan_index)
            date = time2datetime(matching_records[0])

        # Plot FOV outline
        stid = RadarID(dmap_data[0]['stid'])
        if ranges == [] or ranges is None:
            try:
                # If not given, get ranges from data file
                ranges = [0, dmap_data[0]['nrang']]
            except KeyError:
                # Otherwise, default to [0,75]
                ranges = [0, SuperDARNRadars.radars[stid].range_gate_45]

        # Get rsep and frang from data unless not there then take defaults
        # of 180 km for frang and 45 km for rsep as these are most commonly
        # used
        try:
            frang = dmap_data[0]['frang']
        except KeyError:
            frang = 180

        try:
            rsep = dmap_data[0]['rsep']
        except KeyError:
            rsep = 45

        if coords != Coords.GEOGRAPHIC and projs == Projs.GEO:
            raise plot_exceptions.NotImplemented("AACGM coordinates are"
                                                 " not implemented for "
                                                 " geographic projections"
                                                 " right now, if you would"
                                                 " like to see it sooner"
                                                 " please help out at "
                                                 "https://github.com"
                                                 "/SuperDARN/pyDARN")

        beam_corners_lats, beam_corners_lons =\
            coords(stid=RadarID(dmap_data[0]['stid']), rsep=rsep, frang=frang,
                   gates=ranges, date=date, **kwargs)

        fan_shape = beam_corners_lons.shape
        if ranges[0] < ranges[1] - fan_shape[0]:
            ranges[0] = ranges[1] - fan_shape[0] + 1
        rs = beam_corners_lats
        if projs == Projs.POLAR:
            thetas = np.radians(beam_corners_lons)
        else:
            thetas = beam_corners_lons

        # Get range-gate data and groundscatter array for given scan
        # fan_shape has no -1 as when given ranges we want to include the
        # both ends for the ranges given
        scan = np.zeros((fan_shape[0], fan_shape[1]-1))
        grndsct = np.zeros((fan_shape[0], fan_shape[1]-1))
        # Colour table and max value selection depending on parameter plotted
        # Load defaults if none given
        if cmap is None:
            cmap = {'p_l': PyDARNColormaps.PYDARN_PLASMA,
                    'v': PyDARNColormaps.PYDARN_VELOCITY,
                    'w_l': PyDARNColormaps.PYDARN_VIRIDIS,
                    'elv': PyDARNColormaps.PYDARN_INFERNO}
            cmap = cmap[parameter]
        else:
            cmap = cm.get_cmap(cmap)

        # Set background to transparent - avoids carry over
        # does not interfere with the fov color if chosen
        cmap.set_bad(alpha=0.0)

        # Setting zmin and zmax
        defaultzminmax = {'p_l': [0, 50], 'v': [-200, 200],
                          'w_l': [0, 250], 'elv': [0, 50]}
        if zmin is None:
            zmin = defaultzminmax[parameter][0]
        if zmax is None:
            zmax = defaultzminmax[parameter][1]
        norm = colors.Normalize
        norm = norm(zmin, zmax)

        for rec in matching_records:
            try:
                # get a list of gates where there is data
                slist = rec['slist']
                # get the beam number for the record
                beami = rec['bmnum']

                # Exclude ranges larger than the expected maximum.
                # This is a temporary fix to manage inconsistencies between the
                # fitacf files and the hardware files. The issue will be
                # fully resolved when the `rpos` code is committed.
                good_data = np.where((slist >= ranges[0]) &
                                     (slist < ranges[1]))
                slist = slist[good_data]
                temp_data = rec[parameter][good_data]
                temp_ground = rec['gflg'][good_data]

                scan[slist-ranges[0], beami] = temp_data
                grndsct[slist-ranges[0], beami] = temp_ground
            # if there is no slist field this means partial record
            except KeyError:
                partial_record_warning()
                continue

        # Begin plotting by iterating over ranges and beams
        if beam is not None:
            thetas = thetas[0:ranges[1]-ranges[0]+1, beam:beam+2]
            rs = rs[0:ranges[1]-ranges[0]+1, beam:beam+2]
            scan = scan[0:ranges[1]-ranges[0], beam:beam+1]
            grndsct = grndsct[0:ranges[1]-ranges[0], beam:beam+1]
        else:
            thetas = thetas[0:ranges[1]-ranges[0]+1]
            rs = rs[0:ranges[1]-ranges[0]+1]
            scan = scan[0:ranges[1]-ranges[0]]
            grndsct = grndsct[0:ranges[1]-ranges[0]]

        # Set up axes in correct hemisphere
        stid = RadarID(dmap_data[0]['stid'])
        kwargs['hemisphere'] = SuperDARNRadars.radars[stid].hemisphere

        ax, ccrs = projs(date=date, ax=ax, **kwargs)

        if ccrs is None:
            transform = ax.transData
        else:
            if ball_and_stick:
                transform = ccrs.Geodetic()
            else:
                transform = ccrs.PlateCarree()

        if not ball_and_stick:
            ax.pcolormesh(thetas, rs,
                          np.ma.masked_array(scan, ~scan.astype(bool)),
                          norm=norm, cmap=cmap, transform=transform,
                          zorder=2)
        else:
            # Get center of each gate instead of edges
            fan_shape = thetas.shape
            for i in range(0, fan_shape[0]-1):
                for j in range(0, fan_shape[1]-1):
                    t_center = (thetas[i, j] + thetas[i+1, j]
                                + thetas[i, j+1] + thetas[i+1, j+1]) / 4
                    r_center = (rs[i, j] + rs[i+1, j]
                                + rs[i, j+1] + rs[i+1, j+1]) / 4
                    if scan[i, j] != 0.0:
                        col = cmap((scan[i, j] - zmin) / (zmax-zmin))
                        # Plot balls!
                        ax.scatter(t_center, r_center, color=col, s=1.0,
                                   transform=transform, zorder=3.0)
                        # Stick only needed for velocity data
                        if parameter == 'v':
                            # Get azimuth in correct coord system
                            if projs == Projs.POLAR:
                                lat = r_center
                                lon = np.degrees(t_center)
                            else:
                                lat = r_center
                                lon = t_center
                            azm = Fan.get_gate_azm(lon, lat, stid,
                                                   coords, date)

                            # Make sure each coordinate is in correct
                            # units again
                            thetas_calc = np.radians(lon)
                            rs_calc = lat

                            hemisphere = SuperDARNRadars.radars[stid]\
                                                        .hemisphere

                            # Find the end point of the stick to plot
                            # Angle to rotate each vector
                            alpha = thetas_calc

                            # Convert to Cartesian
                            start_pos_x = (90 - abs(rs_calc)) \
                                * np.cos(thetas_calc)
                            start_pos_y = (90 - abs(rs_calc)) \
                                * np.sin(thetas_calc)

                            # Results LOS vector in x and y
                            los_x = -scan[i, j] * np.cos(
                                    np.radians(-azm * hemisphere.value))
                            los_y = -scan[i, j] * np.sin(
                                    np.radians(-azm * hemisphere.value))

                            # Rotate vector into same ref frame
                            vec_x = (los_x * np.cos(alpha)) \
                                - (los_y * np.sin(alpha))
                            vec_y = (los_x * np.sin(alpha)) \
                                + (los_y * np.cos(alpha))

                            # New vector end points
                            end_pos_x = start_pos_x\
                                + (vec_x * hemisphere.value / len_factor)
                            end_pos_y = start_pos_y\
                                + (vec_y * hemisphere.value / len_factor)
                            # Convert back to polar for plotting
                            end_rs = 90 - (np.sqrt(end_pos_x**2
                                                   + end_pos_y**2))
                            end_thetas = np.arctan2(end_pos_y, end_pos_x)
                            end_rs = end_rs * hemisphere.value

                            # Convert to degrees for geo/mag plots
                            if projs != Projs.POLAR:
                                end_thetas = np.degrees(end_thetas)
                            # Plot sticks!
                            ax.plot([t_center, end_thetas],
                                     [r_center, end_rs],
                                     color=col, zorder=3.0, linewidth=0.5,
                                     transform=transform)

                    # Plot ground scatter balls (no sticks)
                    if groundscatter and grndsct[i, j] != 0.0:
                        ax.scatter(t_center, r_center, c='grey', s=1.0,
                                   transform=transform, zorder=3.0)

        # plot the groundscatter as grey fill
        if groundscatter and not ball_and_stick:
            gs_color = colors.ListedColormap(['grey'])
            ax.pcolormesh(thetas, rs,
                          np.ma.masked_array(grndsct,
                                             ~grndsct.astype(bool)),
                          cmap=gs_color,
                          transform=transform, zorder=3)
        if ccrs is None:
            azm = np.linspace(0, 2 * np.pi, 100)
            r, th = np.meshgrid(rs, azm)
            ax.plot(azm, r, color='k', ls='none')
            ax.grid(True)

        if boundary:
            Fan.plot_fov(stid=RadarID(dmap_data[0]['stid']), date=date, ax=ax,
                         ccrs=ccrs, coords=coords, projs=projs, rsep=rsep,
                         frang=frang, ranges=ranges, beam=beam, **kwargs)

        # Create color bar if True
        if colorbar is True:
            mappable = cm.ScalarMappable(norm=norm, cmap=cmap)
            locator = ticker.MaxNLocator(symmetric=True, min_n_ticks=3,
                                         integer=True, nbins='auto')
            ticks = locator.tick_values(vmin=zmin, vmax=zmax)

            if zmin == 0:
                extend = 'max'
            else:
                extend = 'both'

            if cax is None:
                cax = ax.inset_axes([1.1, 0.0, 0.05, 1.0])
            cb = ax.figure.colorbar(mappable, ax=ax, cax=cax, extend=extend,
                                    ticks=ticks)

            if colorbar_label != '':
                cb.set_label(colorbar_label)
        else:
            cb = None

        start_time = time2datetime(matching_records[0])
        if title:
            end_time = time2datetime(matching_records[-1])
            title = Fan.__add_title__(start_time, end_time)
            ax.set_title(title)

        if determine_embargo(start_time,
                             matching_records[0]['cp'],
                             SuperDARNRadars.radars[
                                RadarID(matching_records[0]['stid'])].name):
            add_embargo(plt.gcf())

        return {'ax': ax,
                'ccrs': ccrs,
                'cm': cmap,
                'cb': cb,
                'fig': plt.gcf(),
                'data': {'beam_corners_lats': beam_corners_lats,
                         'beam_corners_lons': beam_corners_lons,
                         'scan_data': scan,
                         'ground_scatter': grndsct}
                }


    @staticmethod
    def plot_fan_input(data_array: list = [], data_datetime: dt.datetime = [],
                       ax: object = None, stid: RadarID = None, data_groundscatter: list = [],
                       rsep: int = 45, frang: int = 180,
                       data_parameter: str = 'v', cmap: str = None, zmin: int = None,
                       zmax: int = None, colorbar: bool = True,
                       colorbar_label: str = '', cax=None, boundary: bool = True,
                       projs: Projs = Projs.POLAR,
                       coords: Coords = Coords.AACGM_MLT, **kwargs):
        """
        Plots a radar's Field Of View (FOV) fan plot for the given data and
        scan number

        Parameters
        -----------
            data_array: List[ranges, beams]
                Array of data, must be in shape of standard fan plot
            data_datetime: datetime object
                Time at which data is taken or wanted to be plotted
            ax: axes.Axes
                Pre-defined axis object to pass in, must currently be
                polar projection
                Default: Generates a polar projection for the user
                with MLT/latitude labels
            data_groundscatter: list[beams, ranges]
                Boolean array of same size which denotes groundscatter
            rsep: int
                Separation between range gates, in kilometers.
                Default: 45
            frang: int
                Kilometers to first range.
                Default: 180
            stid: RadarID
                StationID of the radar of interest
            data_parameter: str
                Key name indicating which parameter to plot.
                Default: v (Velocity). Alternatives: 'p_l', 'w_l', 'elv'
                Used to grab default colourmaps
            cmap: matplotlib.cm
                matplotlib colour map
                https://matplotlib.org/tutorials/colors/colormaps.html
                Default: Official pyDARN colour map for given parameter
            zmin: int
                The minimum parameter value for coloring
                Default: {'p_l': [0], 'v': [-200], 'w_l': [0], 'elv': [0]}
            zmax: int
                The maximum parameter value for  coloring
                Default: {'p_l': [50], 'v': [200], 'w_l': [250], 'elv': [50]}
            colorbar: bool
                Draw a colourbar if True
                Default: True
            colorbar_label: str
                the label that appears next to the colour bar.
                Requires colorbar to be true
                Default: ''
            cax: axes.Axes
                Pre-defined axis for the colorbar. If not specified and colorbar
                is True, a new axis will be created.
            boundary: bool
                if true then plots the FOV boundaries
                default: true
            projs: Enum
                choice of projection for plot
                default: Projs.POLAR (polar projection)
            coords: Enum
                choice of plotting coordinates
                default: Coords.AACGM_MLT (Magnetic Lat and MLT)
            kwargs: key = value
                Additional keyword arguments to be used in projection plotting
                and plot_fov for possible keywords, see: projections.axis_polar

        Returns
        -----------
        beam_corners_aacgm_lats
            n_beams x n_gates numpy array of latitudes
            return values dependent on given coords enum
        beam_corners_aacgm_lons
            n_beams x n_gates numpy array of longitudes or MLT
            return values dependent on given coords enum

        See Also
        --------
            plot_fov
        """
        # Checks on data before plotting
        stid_beams = SuperDARNRadars.radars[stid].hardware_info.beams
        if stid_beams != len(data_array[0]):
            warnings.warn('The data you have inputted to the method does not '
                          'match the expected number of beams for this radar. '
                          'The data will still plot, but the position of the '
                          'extra beams may not be as expected.')

        beam_corners_lats, beam_corners_lons =\
            coords(stid=stid, rsep=rsep, frang=frang,
                   beams=len(data_array[0]),
                   gates=[0, len(data_array)], date=data_datetime,
                   **kwargs)

        rs = beam_corners_lats
        if projs == Projs.POLAR:
            thetas = np.radians(beam_corners_lons)
        else:
            thetas = beam_corners_lons

        scan = np.ma.array(data_array, mask=np.isnan(data_array))
        grndsct = np.array(data_groundscatter)

        # Colour table and max value selection depending on parameter plotted
        # Load defaults if none given
        if cmap is None:
            cmap = {'p_l': PyDARNColormaps.PYDARN_PLASMA,
                    'v': PyDARNColormaps.PYDARN_VELOCITY,
                    'w_l': PyDARNColormaps.PYDARN_VIRIDIS,
                    'elv': PyDARNColormaps.PYDARN_INFERNO}
            cmap = cmap[data_parameter]
        else:
            cmap = cm.get_cmap(cmap)

        # Set background to transparent - avoids carry over
        # does not interfere with the fov color if chosen
        cmap.set_bad(alpha=0.0)

        # Setting zmin and zmax
        defaultzminmax = {'p_l': [0, 50], 'v': [-200, 200],
                          'w_l': [0, 250], 'elv': [0, 50]}
        if zmin is None:
            zmin = defaultzminmax[data_parameter][0]
        if zmax is None:
            zmax = defaultzminmax[data_parameter][1]
        norm = colors.Normalize
        norm = norm(zmin, zmax)

        kwargs['hemisphere'] = SuperDARNRadars.radars[stid].hemisphere
        ax, ccrs = projs(date=data_datetime, ax=ax, **kwargs)

        if ccrs is None:
            transform = ax.transData
        else:
            transform = ccrs.PlateCarree()

        # Plot the data in the scan
        ax.pcolormesh(thetas, rs, scan,
                      norm=norm, cmap=cmap, transform=transform,
                      zorder=2)
        # Plot the groundscatter as grey fill
        if data_groundscatter != []:
            gs_color = colors.ListedColormap(['grey'])
            ax.pcolormesh(thetas, rs,
                          np.ma.masked_array(grndsct,
                                             ~grndsct.astype(bool)),
                          cmap=gs_color,
                          transform=transform, zorder=3)
        if ccrs is None:
            azm = np.linspace(0, 2 * np.pi, 100)
            r, th = np.meshgrid(rs, azm)
            ax.plot(azm, r, color='k', ls='none')
            ax.grid(True)

        if boundary:
            Fan.plot_fov(stid=stid, date=data_datetime, ax=ax,
                         ccrs=ccrs, coords=coords, projs=projs, rsep=rsep,
                         frang=frang, ranges=[0, len(data_array)], **kwargs)

        # Create color bar if True
        if colorbar is True:
            mappable = cm.ScalarMappable(norm=norm, cmap=cmap)
            locator = ticker.MaxNLocator(symmetric=True, min_n_ticks=3,
                                         integer=True, nbins='auto')
            ticks = locator.tick_values(vmin=zmin, vmax=zmax)

            if zmin == 0:
                extend = 'max'
            else:
                extend = 'both'

            if cax is None:
                cax = ax.inset_axes([1.1, 0.0, 0.05, 1.0])
            cb = ax.figure.colorbar(mappable, ax=ax, cax=cax, extend=extend,
                                    ticks=ticks)

            if colorbar_label != '':
                cb.set_label(colorbar_label)
        else:
            cb = None

        return {'ax': ax,
                'ccrs': ccrs,
                'cm': cmap,
                'cb': cb,
                'fig': plt.gcf(),
                'data': {'beam_corners_lats': beam_corners_lats,
                         'beam_corners_lons': beam_corners_lons,
                         'scan_data': scan,
                         'ground_scatter': grndsct}
                }

    @staticmethod
    def plot_fov(stid: RadarID, date: dt.datetime,
                 ax=None, ccrs=None, ranges: List = None, boundary: bool = True,
                 rsep: int = 45, frang: int = 180,
                 projs: Projs = Projs.POLAR,
                 coords: Coords = Coords.AACGM_MLT,
                 fov_color: str = None, alpha: float = 0.5,
                 radar_location: bool = True, radar_label: bool = False,
                 line_color: str = 'black',
                 grid: bool = False, beam: int = None,
                 line_alpha: float = 0.5, **kwargs):
        """
        plots only the field of view (FOV) for a given radar station ID (stid)

        Parameters
        -----------
            stid: RadarID
                Radar station ID
            ax: matplotlib.axes.Axes
                Pre-defined axis object to pass in.
            ccrs:
            date: datetime object
                Sets the datetime used to find the coordinates of the FOV
                Default: Current time
            ranges: list
                Set to a two element list of the lower and upper ranges to plot
                If None, the  max will be obtained by SuperDARNRadars
                Default: None
            boundary: bool
                Set to false to not plot the outline of the FOV
                Default: True
            rsep: int
                Separation between range gates, in kilometers.
                Default: 45
            frang: int
                Kilometers to first range.
                Default: 180
            projs: Projs object
                Sets the projection type for the plot
                Default: Projs.POLAR
            coords: Coords object

            grid: bool
                Set to false to not plot the grid of gates in the FOV
                Default: False
            fov_color: str
                fov color to fill in the boundary
                default: None
            line_color: str
                line color of the fov plot
                default: black
            alpha: int
                alpha controls the transparency of
                the fov color
                Default: 0.5
            line_alpha: int
                line_alpha controls the transparency of
                the boundary and grid lines of the fov
                Default: 0.5
            beam : int or None
                integer indicating if the user would like to plot a single beam
            radar_location: bool
                Add a dot where radar is located if True
                Default: False
            radar_label: bool
                Add a label with the radar abbreviation if True
                Default: False
            kwargs: key = value
                Additional keyword arguments to be used in projection plotting
                For possible keywords, see: projections.axis_polar

        Returns
        -------
            beam_corners_aacgm_lats - list of beam corners AACGM latitudes
            beam_corners_aacgm_lons - list of beam corners AACGM longitudes
            beam_corners_lon - theta polar coordinates
            rs - radius polar coordinates
        """
        if ranges == [] or ranges is None:
            ranges = [0, SuperDARNRadars.radars[stid].range_gate_45]

        if not date:
            date = dt.datetime.now()

        # Get radar beam/gate locations
        beam_corners_lats, beam_corners_lons = \
            coords(stid=stid, gates=ranges, rsep=rsep, frang=frang,
                   date=date, **kwargs)

        # If beam selected then reduce lats and lons array
        if beam is not None:
            beam_corners_lats = beam_corners_lats[:, beam:beam+2]
            beam_corners_lons = beam_corners_lons[:, beam:beam+2]

        if projs == Projs.POLAR:
            beam_corners_lons = np.radians(beam_corners_lons)

        # This section corrects winding order for cartopy plots on a sphere
        # so that the outline is always anti-clockwise and will fill inside
        bmsep = SuperDARNRadars.radars[stid].hardware_info.beam_separation
        if projs != Projs.POLAR and bmsep < 0:
            beam_corners_lons = beam_corners_lons[::-1]
            beam_corners_lats = beam_corners_lats[::-1]

        # Setup plot
        # This may screw up references
        hemisphere = SuperDARNRadars.radars[stid].hemisphere
        if ax is None:
            # Get the hemisphere to pass to plotting projection
            kwargs['hemisphere'] = hemisphere
            ax, ccrs = projs(date=date, **kwargs)
        if ccrs is None:
            transform = ax.transData
        else:
            transform = ccrs.Geodetic()

        if boundary:
            # left boundary line
            ax.plot(beam_corners_lons[0:ranges[1]-ranges[0]+1, 0],
                    beam_corners_lats[0:ranges[1]-ranges[0]+1, 0],
                    color=line_color, linewidth=0.5,
                    alpha=line_alpha, transform=transform, zorder=3)
            # top radar arc
            ax.plot(beam_corners_lons[ranges[1]-ranges[0],
                                      0:beam_corners_lons.shape[1]],
                    beam_corners_lats[ranges[1]-ranges[0],
                                      0:beam_corners_lons.shape[1]],
                    color=line_color, linewidth=0.5,
                    alpha=line_alpha, transform=transform, zorder=3)
            # right boundary line
            ax.plot(beam_corners_lons[0:ranges[1]-ranges[0]+1,
                                      beam_corners_lons.shape[1] - 1],
                    beam_corners_lats[0:ranges[1]-ranges[0]+1,
                                      beam_corners_lons.shape[1] - 1],
                    color=line_color, linewidth=0.5,
                    alpha=line_alpha, transform=transform, zorder=3)
            # bottom arc
            ax.plot(beam_corners_lons[0, 0:beam_corners_lons.shape[1]],
                    beam_corners_lats[0, 0:beam_corners_lons.shape[1]],
                    color=line_color, linewidth=0.5, alpha=line_alpha,
                    transform=transform, zorder=3)

        fan_shape = beam_corners_lons.shape

        if grid:
            # This plots lines along the beams
            for bm in range(fan_shape[1]):
                ax.plot(beam_corners_lons[0:ranges[1] + 1, bm - 1],
                        beam_corners_lats[0:ranges[1] + 1, bm - 1],
                        color=line_color, linewidth=0.2,
                        alpha=line_alpha, transform=transform,
                        zorder=3)
            # This plots arcs along the gates
            for g in range(ranges[1] - ranges[0] + 1):
                ax.plot(beam_corners_lons[g - 1,
                                          0:beam_corners_lons.shape[1]],
                        beam_corners_lats[g - 1,
                                          0:beam_corners_lons.shape[1]],
                        color=line_color, linewidth=0.2,
                        alpha=line_alpha, transform=transform,
                        zorder=3)

        if radar_location:
            Fan.plot_radar_position(stid, ax, date=date, line_color=line_color,
                                    transform=transform, projs=projs,
                                    coords=coords, ccrs=ccrs, **kwargs)
        if radar_label:
            Fan.plot_radar_label(stid, ax, date=date, line_color=line_color,
                                 transform=transform, projs=projs,
                                 coords=coords, ccrs=ccrs, **kwargs)

        if fov_color is not None:
            theta = beam_corners_lons[0:ranges[1] + 1, 0]
            theta =\
                np.append(theta,
                          beam_corners_lons[ranges[1]-ranges[0],
                                            0:beam_corners_lons.shape[1]-1])
            theta =\
                np.append(theta,
                          np.flip(beam_corners_lons[0:ranges[1] -
                                                    ranges[0] + 1,
                                                    beam_corners_lons.
                                                    shape[1] - 1]))
            theta =\
                np.append(theta,
                          np.flip(beam_corners_lons[0,
                                                    0:beam_corners_lons.
                                                    shape[1] - 1]))

            r = beam_corners_lats[0:ranges[1] + 1, 0]
            r =\
                np.append(r, beam_corners_lats[ranges[1]-ranges[0],
                                               0:beam_corners_lons.
                                               shape[1] - 1])
            r =\
                np.append(r,
                          np.flip(beam_corners_lats[0:ranges[1] - ranges[0]+1,
                                                    beam_corners_lons.
                                                    shape[1] - 1]))
            r =\
                np.append(r,
                          np.flip(beam_corners_lats[0,
                                                    0:beam_corners_lons.
                                                    shape[1] - 1]))

            theta = np.flip(theta)
            r = np.flip(r)

            ax.fill(theta, r, color=fov_color, alpha=alpha, zorder=1,
                    transform=transform)

        return {'ax': ax,
                'ccrs': ccrs,
                'cm': None,
                'cb': None,
                'fig': plt.gcf(),
                'data': {'beam_corners_lats': beam_corners_lats,
                         'beam_corners_lons': beam_corners_lons}
                }

    @staticmethod
    def get_gate_azm(theta: float, r: float, stid: RadarID, coords, date):
        """
        gets the azimuth of the gate, requires some changes depending on
        coordinates before using calculate_azimuth

        Parameters
        ----------
            theta: float
                longitude
            r: float
                latitude
            stid: RadarID
                station id of radar
            coords: Enum
                enumeration of coordinate system
            date: datetime object
                date of data

        Returns
        -------
            azm: float
                azimuth direction of radar from gate in coordinate system
                given
        """
        # Get position of radar in geographic from hdw files
        radlat = SuperDARNRadars.radars[stid].hardware_info.geographic.lat
        radlon = SuperDARNRadars.radars[stid].hardware_info.geographic.lon
        # Convert radar position to correct coordinate system
        if coords == Coords.AACGM_MLT or coords == Coords.AACGM:
            geomag_radar = aacgmv2.get_aacgm_coord(radlat, radlon, 250, date)
            radlat = geomag_radar[0]
            radlon = geomag_radar[1]
            if coords == Coords.AACGM_MLT:
                mltshift = geomag_radar[1] -\
                        (aacgmv2.convert_mlt(geomag_radar[1], date) * 15)
                radlon = geomag_radar[1] - mltshift[0]
        # Call calculate azimuth function from geo
        azm = calculate_azimuth(r, theta, 300, radlat, radlon, 300)
        return azm

    @staticmethod
    def plot_radar_position(stid: RadarID, ax: axes.Axes,
                            date: dt.datetime,
                            transform: object = None,
                            coords: Coords = Coords.AACGM_MLT,
                            projs: Projs = Projs.POLAR,
                            line_color: str = 'black',
                            marker: str = '.',
                            markersize: int = 5,
                            **kwargs):
        """
        Plots a symbol at the position of a given radar station ID (stid)

        Parameters
        -----------
            stid: RadarID
                Radar station ID
            ax: matplotlib.axes.Axes
                Pre-defined axis object to plot on.
            date: datetime.datetime
                Sets the datetime used to find the coordinates of the
                FOV
            transform:
            coords: Coords object
            projs: Projs object
            line_color: str
                Color of the symbol
                Default: black
            marker: str
                Controls which symbol is plotted.
                Default: "."
                See https://matplotlib.org/stable/api/markers_api.html#module-matplotlib.markers for options
            markersize: int
                Controls the size of the symbol plotted, "s" passed to ax.scatter().
                Default: 5

        Returns
        -------
            No variables returned
        """
        # Get location of radar
        lat = SuperDARNRadars.radars[stid].hardware_info.geographic.lat
        lon = SuperDARNRadars.radars[stid].hardware_info.geographic.lon
        # Convert to geomag coords
        if coords == Coords.AACGM_MLT or coords == Coords.AACGM:
            geomag_radar = aacgmv2.get_aacgm_coord(lat, lon, 250, date)
            lat = geomag_radar[0]
            lon = geomag_radar[1]
            if coords == Coords.AACGM_MLT:
                mltshift = geomag_radar[1] -\
                        (aacgmv2.convert_mlt(geomag_radar[1], date) * 15)
                lon = geomag_radar[1] - mltshift
        if projs == Projs.POLAR:
            lon = np.radians(lon)
        # Plot a dot at the radar site
        ax.scatter(lon, lat, c=line_color, s=markersize, transform=transform, marker=marker)
        return

    @staticmethod
    def plot_radar_label(stid: RadarID, ax: axes.Axes,
                         date: dt.datetime,
                         coords: Coords = Coords.AACGM_MLT,
                         projs: Projs = Projs.POLAR,
                         line_color: str = 'black', transform: object = None,
                         **kwargs):
        """
        plots only string at the position of a given radar station ID (stid)

        Parameters
        -----------
            stid: RadarID
                Radar station ID
            ax: matplotlib.axes.Axes
                Pre-defined axis object to plot on.
            coords: Coords object
            projs: Projs object
            date: datetime.datetime object
                sets the datetime used to find the coordinates of the
                FOV
            line_color: str
                color of the text
                default: black
            transform:

        Returns
        -------
            No variables returned
        """
        if coords == Coords.AACGM_MLT or coords == Coords.AACGM:
            lat, lon = SuperDARNRadars.radars[stid].mag_label
        else:
            lat, lon = SuperDARNRadars.radars[stid].geo_label

        # Label text
        label_str = ' ' + SuperDARNRadars.radars[stid]\
                    .hardware_info.abbrev.upper()

        # Convert to geomag coords
        if coords == Coords.AACGM_MLT:
            mltshift = lon -\
                    (aacgmv2.convert_mlt(lon, date) * 15)
            lon = lon - mltshift
        if projs == Projs.POLAR:
            lon = np.radians(lon)

        theta_text = lon
        r_text = lat

        ax.text(theta_text, r_text, label_str, ha='center',
                transform=transform, c=line_color)
        return

    @staticmethod
    def __add_title__(first_timestamp: dt.datetime,
                      end_timestamp: dt.datetime):
        title = "{year}-{month}-{day} {start_hour}:{start_minute}:{second} -"\
                " {end_hour}:{end_minute}:{end_second}"\
                "".format(year=first_timestamp.year,
                          month=str(first_timestamp.month).zfill(2),
                          day=str(first_timestamp.day).zfill(2),
                          start_hour=str(first_timestamp.hour).zfill(2),
                          start_minute=str(first_timestamp.minute).zfill(2),
                          second=str(first_timestamp.second).zfill(2),
                          end_hour=str(end_timestamp.hour).
                          zfill(2),
                          end_minute=str(end_timestamp.minute).
                          zfill(2),
                          end_second=str(end_timestamp.second).zfill(2))
        return title



#  Copyright (C) 2020 SuperDARN Canada, University of Saskatchewan
#  Author: Cooper Ross Robertson, Summer Student 2020, Marina Schmidt
#
# Modifications:
# 2023-06-28 Carley Martin refactored return values
#
# Disclaimer:
# pyDARN is under the LGPL v3 license found in the root directory LICENSE.md
# Everyone is permitted to copy and distribute verbatim copies of this license
# document, but changing it is not allowed.
#
# This version of the GNU Lesser General Public License incorporates the terms
# and conditions of version 3 of the GNU General Public License,
# supplemented by the additional permissions listed below.
#

import copy
import matplotlib.pyplot as plt
import numpy as np

from typing import List

from pydarn import SuperDARNRadars, exceptions, RTP, RadarID


class Power():
    """
    Power class to plot SuperDARN RAWACF data power related.


    Methods
    -------
    plot_pwr0_statistic()
    """
    def _str_(self):

        return "This class provides the following methods: \n"\
                " - plot_interference()"

    @classmethod
    def plot_pwr0_statistic(cls, records: List[dict], beam_num: int = 0,
                            min_frequency: float = None,
                            max_frequency: float = None,
                            split_frequency: float = None,
                            statistical_method: object = np.mean):

        """
        This function will calculate and plot a statistic of the lag-0
        power of each record as a function of time.

        This function applies the statistical function (ex. numpy.mean)
        to the pwr0 vector (lag-0 power for each range) for each record
        before plotting the results from all records chronologically.

        Notes
        -----
        This code can be used to study background interference in rawacf data
        when the radar has been operating in a receive-only mode such as
        "politescan" (cpid -3380), or during periods without any obvious
        coherent scatter returns from any range.

        Future Work
        -----------
        Allow for multi-beam comparison to get a directional
        sense of the interference


        Parameters
        ----------
        records: List[dict]
        beam_num: int
            The beam number with the desired data to plot
        min_frequency: float
            set a minimum frequency boundary to plot
            default: None
        max_frequency: float
            set a maximum frequency boundary to plot
            default: None
        split_frequency: float
            frequency to specifically look for or split between depending
            on the setting of compare.
            default: None
        statistical_method: numpy object
            numpy statistical calculation or generic min or max functions
            e.g. numpy.std, numpy.median, numpy.min, numpy.max
            default: numpy.mean

        Raise
        -----
        NoDataFound
        """
        if split_frequency is None:
            if min_frequency is None and max_frequency is None:
                # Plot all frequencies
                records_of_interest = cls.__apply_stat2pwr0(records,
                                                            statistical_method,
                                                            beam_num)
                cls.__plot_pwr0(records_of_interest, beam_num,
                                statistical_method)
                data = {'method': statistical_method,
                        'values': records_of_interest}
            else:
                # plot all frequencies lower than max_frequency
                if min_frequency is None:
                    records_of_interest = cls.\
                            __apply_stat2pwr0(records, statistical_method,
                                              beam_num, '<=', max_frequency)
                    cls.__plot_pwr0(records_of_interest, beam_num,
                                    statistical_method)
                    data = {'method': statistical_method,
                            'values': records_of_interest}

                elif max_frequency is None:
                    # plot all frequencies higher than min_frequency
                    records_of_interest = cls.\
                            __apply_stat2pwr0(records, statistical_method,
                                              beam_num, '>=', min_frequency)
                    cls.__plot_pwr0(records_of_interest, beam_num,
                                    statistical_method)
                    data = {'method': statistical_method,
                            'values': records_of_interest}

                else:
                    # plot all frequencies between min and max frequency
                    # same as single frequency plotting if min=max
                    records_of_interest = cls.\
                        __apply_stat2pwr0(records, statistical_method,
                                          beam_num, '>=', min_frequency)
                    records_of_interest = cls.\
                        __apply_stat2pwr0(records_of_interest,
                                          statistical_method,
                                          beam_num, '<=', max_frequency)

                    cls.__plot_pwr0(records_of_interest, beam_num,
                                    statistical_method)
                    data = {'method': statistical_method,
                            'values': records_of_interest}
        else:
            if min_frequency is None and max_frequency is None:
                # Plot high and low frequencies
                low_frequency_records = cls.\
                    __apply_stat2pwr0(records, statistical_method,
                                      beam_num, '<', split_frequency)
                # arbitrary pick for the inclusion of split_frequency
                high_frequency_records = cls.\
                    __apply_stat2pwr0(records, statistical_method,
                                      beam_num=beam_num, operand='>=',
                                      frequency=split_frequency)
                plt.subplot(2, 1, 1)
                cls.__plot_pwr0(low_frequency_records, beam_num,
                                statistical_method)
                plt.xticks([])
                plt.subplot(2, 1, 2)
                cls.__plot_pwr0(high_frequency_records, beam_num,
                                statistical_method, False)
                data = {'method': statistical_method,
                        'high_freq_values': high_frequency_records,
                        'low_freq_values': low_frequency_records}

            else:
                # plot all frequencies lower than
                # split_frequency and plot all frequencies between
                # split_frequency and max_frequency
                if min_frequency is None:
                    records_of_interest = cls.\
                        __apply_stat2pwr0(records, statistical_method,
                                          beam_num, '<=', max_frequency)
                    low_frequency_records = cls.\
                        __apply_stat2pwr0(records_of_interest,
                                          statistical_method,
                                          beam_num, '<', split_frequency)
                    high_frequency_records = cls.\
                        __apply_stat2pwr0(records_of_interest,
                                          statistical_method,
                                          beam_num, '>=', split_frequency)
                    plt.subplot(2, 1, 1)
                    cls.__plot_pwr0(low_frequency_records, beam_num,
                                    statistical_method)
                    plt.xticks([])
                    plt.subplot(2, 1, 2)
                    cls.__plot_pwr0(high_frequency_records, beam_num,
                                    statistical_method, False)
                    data = {'method': statistical_method,
                            'high_freq_values': high_frequency_records,
                            'low_freq_values': low_frequency_records}

                elif max_frequency is None:
                    # plot all frequencies between min_frequency and
                    # split_frequency and plot all frequencies higher than
                    # split_frequency
                    records_of_interest = cls.\
                        __apply_stat2pwr0(records, statistical_method,
                                          beam_num, '>=', min_frequency)
                    low_frequency_records = cls.\
                        __apply_stat2pwr0(records_of_interest,
                                          statistical_method,
                                          beam_num, '<', split_frequency)
                    high_frequency_records = cls.\
                        __apply_stat2pwr0(records_of_interest,
                                          statistical_method,
                                          beam_num, '>=', split_frequency)
                    plt.subplot(2, 1, 1)
                    cls.__plot_pwr0(low_frequency_records, beam_num,
                                    statistical_method)
                    plt.xticks([])
                    plt.subplot(2, 1, 2)
                    cls.__plot_pwr0(high_frequency_records, beam_num,
                                    statistical_method, False)
                    data = {'method': statistical_method,
                            'high_freq_values': high_frequency_records,
                            'low_freq_values': low_frequency_records}

                # if min=max=split frequency this is the same as
                # split and compare = False
                elif max_frequency != min_frequency:
                    # plot all low and high frequencies between
                    # min and max frequency
                    records_of_interest = cls.\
                        __apply_stat2pwr0(records, statistical_method,
                                          beam_num, '>=', min_frequency)
                    records_of_interest = cls.\
                        __apply_stat2pwr0(records_of_interest,
                                          statistical_method,
                                          beam_num, '<=', max_frequency)

                    low_frequency_records = cls.\
                        __apply_stat2pwr0(records_of_interest,
                                          statistical_method,
                                          beam_num, '<', split_frequency)
                    high_frequency_records = cls.\
                        __apply_stat2pwr0(records_of_interest,
                                          statistical_method,
                                          beam_num, '>=', split_frequency)
                    plt.subplot(2, 1, 1)
                    cls.__plot_pwr0(low_frequency_records, beam_num,
                                    statistical_method)
                    plt.xticks([])
                    plt.subplot(2, 1, 2)
                    cls.__plot_pwr0(high_frequency_records, beam_num,
                                    statistical_method, False)
                    data = {'method': statistical_method,
                            'high_freq_values': high_frequency_records,
                            'low_freq_values': low_frequency_records}
                else:
                    records_of_interest = cls.\
                        __apply_stat2pwr0(records, statistical_method,
                                          beam_num, '==', split_frequency)
                    cls.__plot_pwr0(records_of_interest, beam_num,
                                    statistical_method)
                    data = {'method': statistical_method,
                            'values': records_of_interest}
        return {'ax': plt.gca(),
                'ccrs': None,
                'cm': None,
                'cb': None,
                'fig': plt.gcf(),
                'data': data}

    @staticmethod
    def __plot_pwr0(records: list, beam_num: int, statistical_method: object,
                    title: bool = True):
        """
        This is a method that handles all the plotting calls for pwr0

        Parameters
        ----------
            records: List[dict]
                data records of SuperDARN data
            beam_num: int
                beam number needed for error message if no
                data is found for the frequency
            statistical_method: object
                statistical method to apply to pwr0 array
            title: bool
                to include the title or not
                default: True
        """
        # get station ID and radar abbreviation for the title
        stid = RadarID(records[0]['stid'])
        radar_abbrev = SuperDARNRadars.radars[stid].hardware_info.abbrev
        # plot a time series using RTP class of pwr0
        RTP.plot_time_series(records, parameter='pwr0', beam_num=beam_num)
        plt.ylabel("[{}] A/D power\n [arbitrary units]"
                   "".format(statistical_method.__name__.capitalize()))
        # Find the range of frequencies for the legend
        min_freq = records[0]['tfreq']
        max_freq = records[0]['tfreq']
        for record in records:
            if record['tfreq'] < min_freq:
                min_freq = record['tfreq']
            if record['tfreq'] > max_freq:
                max_freq = record['tfreq']
        if min_freq == max_freq:
            plt.legend(["{} kHz".format(min_freq)])
        else:
            plt.legend(["{}-{} kHz".format(min_freq, max_freq)])

        if title:
            plt.title(' Lag 0 Power for {} Beam: {} '.format(radar_abbrev,
                                                             beam_num))

    @staticmethod
    def __apply_stat2pwr0(records: List[dict], stat_method: object,
                          beam_num: int, operand: str = '',
                          frequency: float = None):
        """
        Finds the records of interest based on the logic expression and
        frequency passed in. The record then is updated with a
        statistical method applying pwr0 to it

        Parameters
        ----------
            records: List[dict]
                data records of SuperDARN data
            stat_method: object
                statistical method to apply to pwr0 array
            beam_num: int
                beam number needed for error message if no
                data is found for the frequency
            operand: str
                str containing a logical operation to
                compare tfreq value in the records to
                the given frequency
            frequency: float
                a number to indicate the frequency threshold
                to compare to the tfreq. If none all
                frequencies are used
                default: None

        Returns
        -------
            records_of_interest: list[dict]
                contains all the records that met the operational
                condition when compared to frequency

        Raises
        ------
            NoDataFound: when no data is found within the comparison
        """

        # deep copy so we don't modify the original records
        records_of_interest = copy.deepcopy(records)

        # tfreq greater than frequency
        if operand == '>':
            records_of_interest = [record for record in records_of_interest
                                   if record['tfreq'] > frequency]
        # tfreq less than frequency
        elif operand == '<':
            records_of_interest = [record for record in records_of_interest
                                   if record['tfreq'] < frequency]
        # tfreq greater than equal to frequency
        elif operand == '>=':
            records_of_interest = [record for record in records_of_interest
                                   if record['tfreq'] >= frequency]
        # tfreq less than equal to frequency
        elif operand == '<=':
            records_of_interest = [record for record in records_of_interest
                                   if record['tfreq'] <= frequency]
        # tfreq equal to frequency
        elif operand == '==':
            records_of_interest = [record for record in records_of_interest
                                   if record['tfreq'] == frequency]

        # if no data is found so raise an error!
        if len(records_of_interest) == 0:
            raise exceptions.plot_exceptions.\
                  NoDataFoundError('tfreq', beam_num,
                                   opt_beam_num=records[0]['bmnum'],
                                   opt_parameter_value=records[0]['tfreq'])

        # loop through the applied records of interest and apply the
        # statistic method provided
        for record in records_of_interest:
            stat_pwr = stat_method(record['pwr0'])
            record.update({'pwr0': stat_pwr})
        return records_of_interest



# Copyright (C) 2021 SuperDARN Canada, University of Saskatchewan
# Author: Marina Schmidt
# Copyright (C) 2012  VT SuperDARN Lab
# Modifications:
# 2022-03-08: MTS - added partial records exception
# 2022-03-18: CJM - Included contour plotting and HMB
# 2022-03-31: CJM - Map info included
# 2022-03-31: CJM - IMF clock angle dial added
# 2022-04-01: CJM - Bug fix for lon shifting to MLT
# 2022-04-28: CJM - Added option to have single color vectors with reference
#                   vector
# 2022-08-15: CJM - Removed plot_FOV call for default uses
# 2022-12-13: CJM - Limited reference vectors to only velocity use
# 2023-06-28: CJM - Refactored return values
# 2024-07-11: CJM - Added potential time series plot
#
# Disclaimer:
# pyDARN is under the LGPL v3 license found in the root directory LICENSE.md
# Everyone is permitted to copy and distribute verbatim copies of this license
# document, but changing it is not allowed.
#
# This version of the GNU Lesser General Public License incorporates the terms
# and conditions of version 3 of the GNU General Public License,
# supplemented by the additional permissions listed below.


"""
Grid plots, mapped to AACGM coordinates in a polar format
"""

import datetime as dt

import matplotlib.axes
import matplotlib.pyplot as plt
import numpy as np
import warnings

from enum import Enum
from matplotlib import ticker, cm, colors
from scipy import special
from typing import List


# Third party libraries
import aacgmv2

from pydarn import (PyDARNColormaps, plot_exceptions, RadarID,
                    standard_warning_format, Re, Hemisphere,
                    time2datetime, find_record, Fan, Projs,
                    MapParams, TimeSeriesParams)
warnings.formatwarning = standard_warning_format


class Maps:
    """
    Maps plots for SuperDARN data

    Methods
    -------
    plot_maps
    calculated_fitted_velocities
    """

    def __str__(self):
        return "This class is static class that provides"\
                " the following methods: \n"\
                "   - plot_maps()\n"

    @classmethod
    def plot_mapdata(cls, dmap_data: List[dict], ax=None,
                     parameter: Enum = MapParams.FITTED_VELOCITY,
                     record: int = 0, start_time: dt.datetime = None,
                     time_delta: float = 1,  alpha: float = 1.0,
                     len_factor: float = 150, color_vectors: bool = True,
                     cmap: str = None, colorbar: bool = True,
                     contour_colorbar: bool = True,
                     colorbar_label: str = '', title: str = '',
                     zmin: float = None, zmax: float = None,
                     hmb: bool = True, boundary: bool = False,
                     radar_location: bool = False, map_info: bool = True,
                     imf_dial: bool = True, reference_vector: int = 500,
                     projs: Projs = Projs.POLAR, **kwargs):
        """
        Plots convection maps data points and vectors

        Parameters
        ----------
            dmap_data : dict[List]
                DMAP data returned from pyDARN.SuperDARNRead or pyDARNio
            ax: object
                matplotlib axis object
            parameters: enum
                enum object to determine what to plot
                default: MapParams.FITTED_VELOCITY
            record: int
                record number to plot
                default: 0
            start_time: datetime.datetime
                datetime object as the start time of the record to plot
                if none then record will be used
                default: none
            time_delta: int
                How close the start_time has to be start_time of the record
                in minutes
                default: 1
            color_vectors: bool
                If True, color the vectors by color map chosen
                If False, color dark grey and include vector reference
                Default: True
            cmap: matplotlib.cm
                matplotlib colour map
                https://matplotlib.org/tutorials/colors/colormaps.html
                Default: Official pyDARN colour map for given parameter
                    MapParams.FITTED_VELOCITY: 'plasma',
                    MapParams.MODEL_VELOCITY: 'plasma',
                    MapParams.RAW_VELOCITY: 'plasma',
                    MapParams.POWER: 'plasma_r',
                    MapParams.SPECTRAL_WIDTH: PyDARNColormaps.PYDARN_VIRIDIS
            zmin: int
                The minimum parameter value for coloring
                Default: MapParams.FITTED_VELOCITY: [0],
                         MapParams.MODEL_VELOCITY: [0],
                         MapParams.RAW_VELOCITY: [0],
                         MapParams.POWER: [0],
                         MapParams.SPECTRAL_WIDTH: [0]
            zmax: int
                The maximum parameter value for  coloring
                Default: MapParams.FITTED_VELOCITY: [1000],
                         MapParams.MODEL_VELOCITY: [1000],
                         MapParams.RAW_VELOCITY: [1000],
                         MapParams.POWER: [250],
                         MapParams.SPECTRAL_WIDTH: [250]
            colorbar: bool
                Draw a colourbar if True
                Default: True
            contour_colorbar: bool
                Draw a contour colourbar if True
                Default: True
            colorbar_label: str
                The label that appears next to the colour bar.
                Requires colorbar to be true
                Default: ''
            title: str
                Adds a title to the plot. If no title is specified,
                one will be provided
                Default: ''
            len_factor: float
                Normalisation factor for the vectors, to control size on plot
                Larger number means smaller vectors on plot
                Default: 150.0
            map_info: bool
                If true, write information about the map on the plot
                (fit order, CPCP, number of points)
            imf_dial: bool
                If True, draw an IMF dial of the magnetic field clock angle.
                Default: True
            reference_vector: int
                If a value is given, a reference velocity vector with
                magnitude of given value, drawn in lower right corner.
                Vector can be turned off by setting this value to False or 0.
                Will not plot for power or spectral width options.
                Default: 500 (vector plotted)
            projs: Enum
                choice of projection for plot
                default: Projs.POLAR (geomagnetic polar projection)
            kwargs: key=value
                uses the parameters for plot_fov and projections.axis

        """
        # Find the record corresponding to the start time
        if start_time is not None:
            record = find_record(dmap_data, start_time, time_delta)
        date = time2datetime(dmap_data[record])

        if cmap is None:
            cmap = {MapParams.FITTED_VELOCITY: PyDARNColormaps.PYDARN_PLASMA_R,
                    MapParams.MODEL_VELOCITY: PyDARNColormaps.PYDARN_PLASMA_R,
                    MapParams.RAW_VELOCITY: PyDARNColormaps.PYDARN_PLASMA_R,
                    MapParams.POWER: PyDARNColormaps.PYDARN_PLASMA,
                    MapParams.SPECTRAL_WIDTH: PyDARNColormaps.PYDARN_VIRIDIS}
            cmap = cmap[parameter]
        # Setting zmin and zmax
        defaultzminmax = {MapParams.FITTED_VELOCITY: [0, 1000],
                          MapParams.MODEL_VELOCITY: [0, 1000],
                          MapParams.RAW_VELOCITY: [0, 1000],
                          MapParams.POWER: [0, 250],
                          MapParams.SPECTRAL_WIDTH: [0, 250]}
        if zmin is None:
            zmin = defaultzminmax[parameter][0]
        if zmax is None:
            zmax = defaultzminmax[parameter][1]

        hemisphere = Hemisphere(dmap_data[record]['hemisphere'])
        norm = colors.Normalize
        norm = norm(zmin, zmax)

        if projs != Projs.POLAR:
            raise plot_exceptions.NotImplemented(" Only polar projections "
                                                 " are implemented for"
                                                 " convection maps."
                                                 " Please set"
                                                 " projs=Projs.POLAR"
                                                 " to plot a convection map.")

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            # If the user wants to plot a FOV boundary or radar location
            # Needs to find the positions for each
            # Else just call the axis maker: proj
            if boundary or radar_location:
                fan_rtn = Fan.plot_fov(RadarID(dmap_data[record]['stid'][0]), date,
                                       ax=ax, boundary=boundary,
                                       radar_location=radar_location,
                                       **kwargs)
                ax = fan_rtn['ax']
                ccrs = fan_rtn['ccrs']
                for stid in dmap_data[record]['stid'][1:]:
                    fan_rtn = Fan.plot_fov(RadarID(stid), date, ax=ax,
                                           boundary=boundary, ccrs=ccrs,
                                           radar_location=radar_location,
                                           **kwargs)
                    ax = fan_rtn['ax']
            else:
                
                ax, ccrs = projs(date, ax=ax, hemisphere=hemisphere, **kwargs)


        if parameter == MapParams.MODEL_VELOCITY:
            try:
                data_lons = dmap_data[record]['model.mlon']
                data_lats = dmap_data[record]['model.mlat']
            except KeyError:
                raise plot_exceptions.PartialRecordsError('model.mlon')
        else:
            try:
                data_lons = dmap_data[record]['vector.mlon']
                data_lats = dmap_data[record]['vector.mlat']
            except KeyError:
                raise plot_exceptions.PartialRecordsError('model.mlat')

        # Arbitrary lon used to calculate the shift required
        shifted_mlts = 0 - (aacgmv2.convert_mlt(0, date) * 15)
        shifted_lons = data_lons - shifted_mlts
        # Note that this "mlons" is adjusted for MLT
        mlons = np.radians(shifted_lons)
        mlats = data_lats

        # If the parameter is velocity then plot the LOS vectors
        # Actual mlons used here, not adjusted mlons (np.radians(data_lons))
        if parameter == MapParams.FITTED_VELOCITY:
            v_mag, azm_v =\
                    cls.calculated_fitted_velocities(mlats=mlats,
                                                     mlons=np.radians(
                                                         data_lons),
                                                     hemisphere=hemisphere,
                                                     fit_coefficient=dmap_data[
                                                         record]['N+2'],
                                                     fit_order=dmap_data[
                                                         record]['fit.order'],
                                                     lat_min=dmap_data[
                                                         record]['latmin'])

        elif parameter == MapParams.MODEL_VELOCITY:
            v_mag = dmap_data[record]['model.vel.median']
            azm_v = np.radians(dmap_data[record]['model.kvect'])
        elif parameter == MapParams.RAW_VELOCITY:
            v_mag = dmap_data[record]['vector.vel.median']
            azm_v = np.radians(dmap_data[record]['vector.kvect'])
        elif parameter == MapParams.POWER:
            v_mag = dmap_data[record]['vector.pwr.median']
            azm_v = np.radians(dmap_data[record]['vector.kvect'])

        elif parameter == MapParams.SPECTRAL_WIDTH:
            v_mag = dmap_data[record]['vector.wdt.median']
            azm_v = np.radians(dmap_data[record]['vector.kvect'])

        if parameter in [MapParams.FITTED_VELOCITY, MapParams.MODEL_VELOCITY,
                         MapParams.RAW_VELOCITY]:
            # Make reference vector and add it to the array to
            # be calculated too
            reflat = (np.abs(plt.gca().get_ylim()[1]) - 5) * hemisphere.value
            reflon = np.radians(45)
            v_mag = np.append(v_mag, reference_vector)
            if hemisphere == Hemisphere.North:
                azm_v = np.append(azm_v, np.radians(135))
            else:
                azm_v = np.append(azm_v, np.radians(45))

            # Angle to "rotate" each vector by to get into same
            # reference frame Controlled by longitude, or "mltitude"
            alpha = np.append(mlons, reflon)
            mlons = np.append(mlons, reflon)
            mlats = np.append(mlats, reflat)

            # Convert initial positions to Cartesian
            start_pos_x = (90 - abs(mlats)) * np.cos(mlons)
            start_pos_y = (90 - abs(mlats)) * np.sin(mlons)

            # Resolve LOS vector in x and y directions,
            # with respect to mag pole
            # Gives zonal and meridional components of LOS vector
            x = -v_mag * np.cos(-azm_v * hemisphere.value)
            y = -v_mag * np.sin(-azm_v * hemisphere.value)

            # Rotate each vector into same reference frame
            # following vector rotation matrix
            # https://en.wikipedia.org/wiki/Rotation_matrix
            vec_x = (x * np.cos(alpha)) - (y * np.sin(alpha))
            vec_y = (x * np.sin(alpha)) + (y * np.cos(alpha))

            # New vector end points, in Cartesian
            end_pos_x = start_pos_x + (vec_x * hemisphere.value / len_factor)
            end_pos_y = start_pos_y + (vec_y * hemisphere.value / len_factor)

            # Convert back to polar for plotting
            end_mlats = 90.0 - (np.sqrt(end_pos_x**2 + end_pos_y**2))
            end_mlons = np.arctan2(end_pos_y, end_pos_x)

            end_mlats = end_mlats * hemisphere.value

            # Plot the vector socks (final vector is the reference
            # vector to be plotted later if required)
            if color_vectors is True:
                for i in range(len(v_mag) - 1):
                    if parameter == MapParams.FITTED_VELOCITY:
                        # Shift HMB lons to MLT
                        shifted_mlts = \
                            dmap_data[record]['boundary.mlon'][0] - \
                            (aacgmv2.convert_mlt(
                                dmap_data[record]['boundary.mlon'][0],
                                date) * 15)
                        hmblons = (dmap_data[record]['boundary.mlon'] -
                                   shifted_mlts) % 360
                        # Find where the closest HMB value is, set equivalent
                        # latitude as the lat limit for plotting
                        rounded_mlon = np.degrees(mlons[i]) % 360
                        ind = (np.abs(hmblons - rounded_mlon)).argmin()
                        lat_limit = dmap_data[record]['boundary.mlat'][ind]
                        if abs(mlats[i]) >= abs(lat_limit):
                            ax.plot([mlons[i], end_mlons[i]],
                                     [mlats[i], end_mlats[i]],
                                     c=cmap(norm(v_mag[i])),
                                     linewidth=0.5, zorder=5.0)
                    else:
                        ax.plot([mlons[i], end_mlons[i]],
                                 [mlats[i], end_mlats[i]],
                                 c=cmap(norm(v_mag[i])),
                                 linewidth=0.5, zorder=5.0)
            else:
                for i in range(len(v_mag) - 1):
                    if parameter == MapParams.FITTED_VELOCITY:
                        # Shift HMB lons to MLT
                        shifted_mlts = \
                            dmap_data[record]['boundary.mlon'][0] - \
                            (aacgmv2.convert_mlt(
                                dmap_data[record]['boundary.mlon'][0],
                                date) * 15)
                        hmblons = (dmap_data[record]['boundary.mlon'] -
                                   shifted_mlts) % 360
                        # Find where the closest HMB value is, set equivalent
                        # latitude as the lat limit for plotting
                        rounded_mlon = np.degrees(mlons[i]) % 360
                        ind = (np.abs(hmblons - rounded_mlon)).argmin()
                        lat_limit = dmap_data[record]['boundary.mlat'][ind]
                        if abs(mlats[i]) >= abs(lat_limit):
                            ax.plot([mlons[i], end_mlons[i]],
                                     [mlats[i], end_mlats[i]], c='#292929',
                                     linewidth=0.5, zorder=5.0)
                    else:
                        ax.plot([mlons[i], end_mlons[i]],
                                 [mlats[i], end_mlats[i]], c='#292929',
                                 linewidth=0.5, zorder=5.0)

        # Plot the sock start dots and reference vector if known
        if color_vectors is True:
            if parameter in [MapParams.MODEL_VELOCITY,
                             MapParams.RAW_VELOCITY]:
                if reference_vector > 0:
                    ax.scatter(mlons[:-1], mlats[:-1], c=v_mag[:-1], s=2.0,
                                vmin=zmin, vmax=zmax,  cmap=cmap, zorder=5.0,
                                clip_on=True)
                    ax.scatter(mlons[-1], mlats[-1], c=v_mag[-1], s=2.0,
                                vmin=zmin, vmax=zmax,  cmap=cmap, zorder=5.0,
                                clip_on=False)
                    ax.plot([mlons[-1], end_mlons[-1]],
                             [mlats[-1], end_mlats[-1]],
                             c=cmap(norm(v_mag[-1])),
                             linewidth=0.5, zorder=5.0, clip_on=False)
                    plt.figtext(0.675, 0.15, str(reference_vector) + ' m/s',
                                fontsize=8)
                else:
                    ax.scatter(mlons[:-1], mlats[:-1], c=v_mag[:-1], s=2.0,
                                vmin=zmin, vmax=zmax,  cmap=cmap, zorder=5.0)
            elif parameter is MapParams.FITTED_VELOCITY:
                # Shift HMB lons to MLT
                shifted_mlts = dmap_data[record]['boundary.mlon'][0] - \
                        (aacgmv2.convert_mlt(
                            dmap_data[record]['boundary.mlon'][0], date) * 15)
                hmblons = (dmap_data[record]['boundary.mlon'] -
                           shifted_mlts) % 360
                # Find where the closest HMB value is, set equivalent
                # latitude as the lat limit for plotting
                for m, mlon in enumerate(mlons[:-1]):
                    rounded_mlon = np.degrees(mlon) % 360
                    ind = (np.abs(hmblons - rounded_mlon)).argmin()
                    lat_limit = dmap_data[record]['boundary.mlat'][ind]
                    if abs(mlats[m]) >= abs(lat_limit):
                        ax.scatter(mlon, mlats[m], color=cmap(norm(v_mag[m])),
                                    s=2.0, zorder=5.0, clip_on=True)
                    else:
                        ax.scatter(mlon, mlats[m], c='#DDDDDD', s=2.0,
                                    zorder=5.0, clip_on=True)
                if reference_vector > 0:
                    ax.scatter(mlons[-1], mlats[-1],
                                color=cmap(norm(v_mag[-1])),
                                s=2.0, zorder=5.0, clip_on=False)
                    ax.plot([mlons[-1], end_mlons[-1]],
                             [mlats[-1], end_mlats[-1]],
                             c=cmap(norm(v_mag[-1])),
                             linewidth=0.5, zorder=5.0, clip_on=False)
                    plt.figtext(0.675, 0.15, str(reference_vector) + ' m/s',
                                fontsize=8)
            # No vector socks on spectral width
            else:
                ax.scatter(mlons[:], mlats[:], c=v_mag[:], s=2.0,
                            vmin=zmin, vmax=zmax,  cmap=cmap, zorder=5.0)

        else:
            # no color so make sure colorbar is turned off
            colorbar = False
            if parameter in [MapParams.MODEL_VELOCITY,
                             MapParams.RAW_VELOCITY]:
                if reference_vector > 0:
                    ax.scatter(mlons[:-1], mlats[:-1], c='#292929', s=2.0,
                                zorder=5.0, clip_on=True)
                    ax.scatter(mlons[-1], mlats[-1], c='#292929', s=2.0,
                                zorder=5.0, clip_on=False)
                    ax.plot([mlons[-1], np.degrees(end_mlons[-1])],
                             [mlats[-1], end_mlats[-1]], c='#292929',
                             linewidth=0.5, zorder=5.0, clip_on=False)
                    plt.figtext(0.675, 0.15, str(reference_vector) + ' m/s',
                                fontsize=8)
                else:
                    ax.scatter(mlons[:-1], mlats[:-1], c='#292929', s=2.0,
                                zorder=5.0)
            elif parameter is MapParams.FITTED_VELOCITY:
                # Shift HMB lons to MLT
                shifted_mlts = dmap_data[record]['boundary.mlon'][0] - \
                        (aacgmv2.convert_mlt(
                            dmap_data[record]['boundary.mlon'][0], date) * 15)
                hmblons = (dmap_data[record]['boundary.mlon'] -
                           shifted_mlts) % 360
                # Find where the closest HMB value is, set equivalent
                # latitude as the lat limit for plotting
                for m, mlon in enumerate(mlons[:-1]):
                    rounded_mlon = np.degrees(mlon) % 360
                    ind = (np.abs(hmblons - rounded_mlon)).argmin()
                    lat_limit = dmap_data[record]['boundary.mlat'][ind]
                    if abs(mlats[m]) >= abs(lat_limit):
                        ax.scatter(mlon, mlats[m], c='#292929', s=2.0,
                                    zorder=5.0, clip_on=True)
                    else:
                        ax.scatter(mlon, mlats[m], c='#DDDDDD', s=2.0,
                                    zorder=5.0, clip_on=True)
                if reference_vector > 0:
                    ax.scatter(mlons[-1], mlats[-1], c='#292929', s=2.0,
                                zorder=5.0, clip_on=False)
                    ax.plot([mlons[-1], end_mlons[-1]],
                             [mlats[-1], end_mlats[-1]], c='#292929',
                             linewidth=0.5, zorder=5.0, clip_on=False)
                    plt.figtext(0.675, 0.15, str(reference_vector) + ' m/s',
                                fontsize=8)
            # No vector socks on spectral width
            else:
                ax.scatter(mlons[:], mlats[:], c='#292929', s=2.0,
                            zorder=5.0)

        if colorbar is True:
            mappable = cm.ScalarMappable(norm=norm, cmap=cmap)
            locator = ticker.MaxNLocator(symmetric=True, min_n_ticks=3,
                                         integer=True, nbins='auto')
            ticks = locator.tick_values(vmin=zmin, vmax=zmax)

            if zmin == 0:
                cb = ax.figure.colorbar(mappable, ax=ax, extend='max',
                                        ticks=ticks)
            else:
                cb = ax.figure.colorbar(mappable, ax=ax, extend='both',
                                        ticks=ticks)

            if colorbar_label != '':
                cb.set_label(colorbar_label)
            else:
                if parameter in [MapParams.FITTED_VELOCITY,
                                 MapParams.MODEL_VELOCITY,
                                 MapParams.RAW_VELOCITY]:
                    cb.set_label('Velocity (m s$^{-1}$)')
                elif parameter is MapParams.SPECTRAL_WIDTH:
                    cb.set_label('Spectral Width (m s$^{-1}$)')
                elif parameter is MapParams.POWER:
                    cb.set_label('Power')
        else:
            cb = None

        # Plot potential contours
        fit_coefficient = dmap_data[record]['N+2']
        fit_order = dmap_data[record]['fit.order']
        lat_shift = dmap_data[record]['lat.shft']
        lon_shift = dmap_data[record]['lon.shft']
        lat_min = dmap_data[record]['latmin']

        _, _, pot_arr, cs, cb_contour = \
            cls.plot_potential_contours(fit_coefficient,
                                        lat_min, date, ax,
                                        lat_shift=lat_shift,
                                        lon_shift=lon_shift,
                                        fit_order=fit_order,
                                        hemisphere=hemisphere,
                                        contour_colorbar=contour_colorbar,
                                        **kwargs)

        if hmb is True:
            # Plot the HMB
            mlats_hmb = dmap_data[record]['boundary.mlat']
            mlons_hmb = dmap_data[record]['boundary.mlon']
            hmb_lon, hmb_lat = cls.plot_heppner_maynard_boundary(mlats_hmb,
                                                                 mlons_hmb,
                                                                 date)
        else:
            hmb_lon = None
            hmb_lat = None

        if title == '':
            title = "{year}-{month}-{day} {start_hour}:{start_minute} -"\
                " {end_hour}:{end_minute}"\
                    "".format(year=date.year,
                              month=str(date.month).zfill(2),
                              day=str(date.day).zfill(2),
                              start_hour=str(date.hour).zfill(2),
                              start_minute=str(date.minute).zfill(2),
                              end_hour=str(dmap_data[record]['end.hour']).
                              zfill(2),
                              end_minute=str(dmap_data[record]['end.minute']).
                              zfill(2))
        plt.title(title)

        model = dmap_data[record]['model.name']
        num_points = len(dmap_data[record]['vector.mlat'])
        pol_cap_pot = dmap_data[record]['pot.drop']
        if map_info is True:
            cls.add_map_info(fit_order, pol_cap_pot, num_points, model)

        bx = dmap_data[record]['IMF.Bx']
        by = dmap_data[record]['IMF.By']
        bz = dmap_data[record]['IMF.Bz']
        delay = dmap_data[record]['IMF.delay']
        bt = np.sqrt(bx**2 + by**2 + bz**2)
        if imf_dial is True:
            # Plot the IMF dial
            cls.plot_imf_dial(ax, by, bz, bt, delay)

        return {'ax': ax,
                'ccrs': None,
                'cm': cmap,
                'cb': [cb, cb_contour],
                'fig': plt.gcf(),
                'data': {'cpcp': pol_cap_pot,
                         'fit_model': model,
                         'fit_order': fit_order,
                         'num_vectors': num_points,
                         'mlats': mlats,
                         'mlons': mlons,
                         'v_mag': v_mag,
                         'azm_v': azm_v,
                         'potential_array': pot_arr,
                         'hmb': [hmb_lat, hmb_lon],
                         'imf': [bx, by, bz, bt, delay],
                         'contours': cs}
                }

    @classmethod
    def index_legendre(cls, el: int, m: int):
        """
        not a 100% how this works some black magic

        parameter
        ---------
            l : int
                doping level
            m : int
                fit order

        return
        ------
            legendre index?
        """
        return (m == 0 and el**2) or ((el != 0)
                                      and (m != 0) and el**2 + 2 * m - 1) or 0

    @classmethod
    def calculated_fitted_velocities(cls, mlats: np.array, mlons: np.array,
                                     fit_coefficient: np.array,
                                     hemisphere: Enum = Hemisphere.North,
                                     fit_order: int = 6, lat_min: int = 60):
        """
        Calculates the fitted velocities using Legrendre polynomial

        Parameters
        ----------
            mlats: Array[float]
                Magnetic Latitude in degrees
            mlons: Array[float]
                Magnetic Longitude in radians
            fit_coefficient: Array[float]
                Value of the mapfile coefficients (from record key ['N+2'])
            hemisphere: int
                1 or -1 for hemisphere North or South
                default: 1 - North
            fit_order: int
                order of the fit
                default: 6
            lat_min: int
                Lower latitude boundary of data in degrees
                default: 60
        """
        # convert earth radius to meters
        Re_meters = Re * 1000.0
        # theta values in radians
        thetas = np.radians(90.0 - abs(mlats))
        thetas_max = np.radians(90.0 - abs(lat_min))

        # Angle to "rotate" each vector by to get into same
        # reference frame Controlled by longitude, or "mltitude"
        alpha = np.pi / thetas_max
        thetas_prime = alpha * thetas
        x = np.cos(thetas_prime)

        # i is the index of the list
        # x_i is the element of x at ith index
        for i, x_i in enumerate(x):
            temp_poly = special.lpmn(fit_order, fit_order, x_i)
            if i == 0:
                legendre_poly = np.append([temp_poly[0]], [temp_poly[0]],
                                          axis=0)
            else:
                legendre_poly = np.append(legendre_poly, [temp_poly[0]],
                                          axis=0)
        legendre_poly = np.delete(legendre_poly, 0, 0)
        phi = mlons

        # now do the index legender part,
        # We are doing Associated Legendre Polynomials but
        # for each polynomial we have two coefficients one
        # for cos(phi) and the other for sin(phi),
        # so we do spherical harmonics for a real valued function using
        # sin(phi) and cos(phi) rather than exp(i*phi).
        # we place an inner function to copying code

        # max index value
        k_max = cls.index_legendre(fit_order, fit_order)

        # set up arrays and small stuff for the E field
        # coefficients calculation
        thetas_ecoeffs = np.zeros((k_max + 2, len(thetas)))
        phi_ecoeffs = np.zeros((k_max + 2, len(thetas)))

        q_prime = np.array(np.where(thetas_prime != 0.0))
        q_prime = q_prime[0]
        q = np.array(np.where(thetas != 0.0))
        q = q[0]

        # finally get to converting coefficients for the potential into
        # coefficients for elec. Field
        fit_coefficient_flat = fit_coefficient.flatten()
        for m in range(fit_order + 1):
            for el in range(m, fit_order + 1):
                k3 = cls.index_legendre(el, m)
                k4 = cls.index_legendre(el, m)

                if k3 >= 0:
                    thetas_ecoeffs[k4, q_prime] =\
                            thetas_ecoeffs[k4, q_prime] -\
                            fit_coefficient_flat[k3] * alpha * el *\
                            np.cos(thetas_prime[q_prime]) / \
                            np.sin(thetas_prime[q_prime]) / Re_meters
                    phi_ecoeffs[k4, q] = phi_ecoeffs[k4, q] - \
                        fit_coefficient_flat[k3 + 1] * m /\
                        np.sin(thetas[q]) / Re_meters
                    phi_ecoeffs[k4 + 1, q] = phi_ecoeffs[k4 + 1, q] + \
                        fit_coefficient_flat[k3] * m /\
                        np.sin(thetas[q]) / Re_meters

                if el < fit_order:
                    k1 = cls.index_legendre(el+1, m)
                else:
                    k1 = -1

                k2 = cls.index_legendre(el, m)

                if k1 >= 0:
                    thetas_ecoeffs[k2, q_prime] =\
                        thetas_ecoeffs[k2, q_prime] + \
                        fit_coefficient_flat[k1] * alpha * (el + 1 + m) / \
                        np.sin(thetas_prime[q_prime]) / Re_meters

                if m > 0:
                    if k3 >= 0:
                        k3 = k3 + 1
                    k4 = k4 + 1

                    if k1 >= 0:
                        k1 = k1 + 1
                    k2 = k2 + 1

                    if k3 >= 0:
                        thetas_ecoeffs[k4, q_prime] =\
                                thetas_ecoeffs[k4, q_prime] \
                                - fit_coefficient_flat[k3] * alpha * el * \
                                np.cos(thetas_prime[q_prime]) / \
                                np.sin(thetas_prime[q_prime]) / Re_meters

                    if k1 >= 0:
                        thetas_ecoeffs[k2, q_prime] = \
                            thetas_ecoeffs[k2, q_prime] \
                            + fit_coefficient_flat[k1] * alpha *\
                            (el + 1 + m) / np.sin(thetas_prime[q_prime]) /\
                            Re_meters

        # Calculate the Electric field positions
        thetas_ecomp = np.zeros(thetas.shape)
        phi_ecomp = np.zeros(thetas.shape)

        for m in range(fit_order + 1):
            for el in range(m, fit_order + 1):
                k = cls.index_legendre(el, m)
                # Now in the IDL code we use
                # legendre_poly[:,l,m] instead of
                # legendre_poly[:,m,l] like here, this is
                # because we have a different
                # organization of legendre_poly due to the
                # way scipy.special.lpmn
                # stores values in arrays...
                if m == 0:
                    thetas_ecomp = thetas_ecomp + thetas_ecoeffs[k, :] * \
                            legendre_poly[:, m, el]
                    phi_ecomp = phi_ecomp + phi_ecoeffs[k, :] * \
                        legendre_poly[:, m, el]
                else:
                    thetas_ecomp = thetas_ecomp + thetas_ecoeffs[k, :] * \
                        legendre_poly[:, m, el] * np.cos(m * phi) + \
                        thetas_ecoeffs[k+1, :] * legendre_poly[:, m, el] * \
                        np.sin(m * phi)
                    phi_ecomp = phi_ecomp + phi_ecoeffs[k, :] * \
                        legendre_poly[:, m, el] * np.cos(m * phi) + \
                        phi_ecoeffs[k+1, :] * legendre_poly[:, m, el] * \
                        np.sin(m * phi)

        # Store the two components of Efield into a single array
        E_field_fit = np.append([thetas_ecomp], [phi_ecomp], axis=0)

        # We'll calculate Bfield magnitude now, need to initialize some more
        # stuff
        # F-region altitude 300 km * 1000 to convert to meteres
        F_altitude = 300.0 * 1000.0
        # dipole earth field in Tesla
        B_field_polar = -0.62e-4
        B_field = B_field_polar * (1.0 - 3.0 * F_altitude / Re_meters) \
            * np.sqrt(3.0 * np.square(np.cos(thetas)) + 1.0) / 2

        # get the velocity components from E-field
        velocity_fit_vectors = np.zeros(E_field_fit.shape)
        velocity_fit_vectors[0, :] = E_field_fit[1, :] / B_field
        velocity_fit_vectors[1, :] = -E_field_fit[0, :] / B_field
        velocity = np.sqrt(np.square(velocity_fit_vectors[0, :]) +
                           np.square(velocity_fit_vectors[1, :]))
        velocity_chk_zero_inds = np.where(velocity != 0.0)
        velocity_chk_zero_inds = velocity_chk_zero_inds[0]

        azm_v = np.zeros(velocity.shape)

        if len(velocity_chk_zero_inds) == 0:
            velocity = np.array([0.0])
            azm_v = np.array([0.0])
        else:
            if hemisphere == Hemisphere.South:
                azm_v[velocity_chk_zero_inds] =\
                        np.arctan2(
                            velocity_fit_vectors[1, velocity_chk_zero_inds],
                            velocity_fit_vectors[0, velocity_chk_zero_inds])
            else:
                azm_v[velocity_chk_zero_inds] =\
                        np.arctan2(
                            velocity_fit_vectors[1, velocity_chk_zero_inds],
                            -velocity_fit_vectors[0, velocity_chk_zero_inds])

        return velocity, azm_v

    @classmethod
    def add_map_info(cls, fit_order: float, pol_cap_pot: float,
                     num_points: float, model: str):
        """
        Annotates the plot with information about the map plotting

        Parameters
        ----------
            ax: object
                matplotlib axis object
            fit_order: int
                order of the fit
            pol_cap_pot: float
                value of the polar cap potential in kV
            num_points: int
                number of vectors plotted
            model: str
                model used to fit data
        """
        text_string = r'$\phi_{PC}$' + ' = ' + str(round(pol_cap_pot/1000))\
                      + ' kV\n' \
                      + 'N = ' + str(num_points) + '\n' \
                      + 'Order: ' + str(fit_order) + '\n' \
                      + 'Model: ' + model + '\n'
        plt.figtext(0.1, 0.1, text_string)

    @classmethod
    def plot_heppner_maynard_boundary(cls, mlats: list, mlons: list,
                                      date: object, line_color: str = 'black',
                                      **kwargs):
        # TODO: No evaluation of coordinate system made! May need if in
        # plotting to plot in radians/geo ect.
        """
        Plots the position of the Heppner-Maynard Boundary

        Parameters
        ----------
            ax: object
                matplotlib axis object
            mlats: List[float]
                Magnetic Latitude in degrees
            mlons: List[float]
                Magnetic Longitude in radians
            date: datetime object
                Date from record
            line_color: str
                Color of the Heppner-Maynard boundary
                Default: black

        """
        # Shift mlon to MLT
        shifted_mlts = mlons[0] - \
            (aacgmv2.convert_mlt(mlons[0], date) * 15)
        shifted_lons = mlons - shifted_mlts
        mlon = np.radians(shifted_lons)

        plt.plot(mlon, mlats, c=line_color, zorder=4.0, **kwargs)
        return mlon, mlats

    @classmethod
    def plot_imf_dial(cls, ax: matplotlib.axes.Axes, by: float = 0, bz: float = 0,
                      bt: float = 0, delay: float = 0):
        """
        Plots an IMF clock angle dial on the existing plot
        Defaults all to 0 if no IMF data available to plot

        Parameters
        ----------
            ax: object
                matplotlib axis object
            by: Float
                Value of the magnetic field in the y-direction (nT)
                Default = 0 nT
            bz: Float
                Value of the magnetic field in the z-direction (nT)
                Default = 0 nT
            bt: Float
                Magnitude of the magnetic field (nT)
                Default = 0 nT
            delay: Float
                Time delay of magnetic field between the
                measuring satellite and the ionosphere (minutes)
                Default = 0 minutes
        """
        # Create new axes inside existing axes
        ax_imf = ax.inset_axes((-0.2, 0.7, 0.4, 0.4))
        ax_imf.axis('off')

        ax_imf.set_xlim([-20.2, 20.2])
        ax_imf.set_ylim([-20.2, 20.2])

        # Plot a Circle
        limit_circle = plt.Circle((0, 0), 10, facecolor='w',
                                  edgecolor='k')
        ax_imf.add_patch(limit_circle)
        # Plot axis lines
        ax_imf.plot([-10, 10], [0, 0], color='k', linewidth=0.5)
        ax_imf.plot([0, 0], [-10, 10], color='k', linewidth=0.5)

        # Plot line for magnetic field
        ax_imf.plot([0, by], [0, bz], color='r')

        # Add axis labels
        ax_imf.annotate('+Z', xy=(-2.5, 11))
        ax_imf.annotate('+Y', xy=(11, -1))

        # Add annotations for delay and Btot
        ax_imf.annotate('|B| = ' + str(round(bt)) + ' nT', xy=(-16, -13),
                        fontsize=7)
        ax_imf.annotate('Delay = -' + str(delay) + ' min', xy=(-16, -17),
                        fontsize=7)

    @classmethod
    def calculate_potentials(cls, fit_coefficient: list, lat_min: list,
                             lat_shift: int = 0, lon_shift: int = 0,
                             fit_order: int = 6, lowlat: int = 30,
                             hemisphere: Enum = Hemisphere.North,
                             **kwargs):
        # TODO: No evaluation of coordinate system made! May need if in
        # plotting to plot in radians/geo ect.
        '''
        Calculates potential across a magnetic lat/lon grid for
        plotting later

        Parameters
        ----------
            fit_coefficient: List[float]
                Value of the coefficient
            lat_min: List[float]
                Minimum latitude that will be evaluated
                Not to be confused with 'lowlat'
            lat_shift: int
                Generic shift in latitude from map file
                default: 0
            lon_shift: int
                Generic shift in longitude from map file
                default: 0
            fit_order: int
                order of the fit
                default: 6
            lowlat: int
                Lowest latitude on plot
                default: 60
            hemisphere: Enum
                Describes the hemisphere, North or South
                default: Hemisphere.North

        '''
        # Lowest latitude to calculate potential to
        theta_max = np.radians(90-np.abs(lat_min)) * hemisphere.value

        # Make a grid of the space the potential is evaluated on
        # in magnetic coordinates
        lat_step = 1
        lon_step = 2
        num_lats = int((90.0 - lowlat) / lat_step) + 1
        num_lons = int(360.0 / lon_step) + 1
        lat_arr = np.array(range(num_lats)) * lat_step + lowlat
        lon_arr = np.array(range(num_lons)) * lon_step

        # Set up Grid
        grid_arr = np.zeros((2, num_lats * num_lons))
        count1 = 0
        for lons in lon_arr:
            for lats in lat_arr:
                grid_arr[0, count1] = lats
                grid_arr[1, count1] = lons
                count1 += 1

        # Convert grid vals to spherical coords
        theta = np.radians(90.0 - np.abs(grid_arr[0, :]))
        phi = np.radians(grid_arr[1, :])

        # Adjusted/Normalised values (runs 0 - pi)
        alpha = np.pi / theta_max
        x = np.cos(alpha*theta)
        # Legendre Polys
        for j, xj in enumerate(x):
            plm_tmp = special.lpmn(fit_order, fit_order, xj)
            if j == 0:
                plm_fit = np.append([plm_tmp[0]], [plm_tmp[0]], axis=0)
            else:
                plm_fit = np.append(plm_fit, [plm_tmp[0]], axis=0)
        # Remove first element as it is duplicated to start off the array
        plm_fit = np.delete(plm_fit, 0, 0)

        # Eval the potential
        lmax = plm_fit.shape
        lmax = lmax[1]
        v = np.zeros(phi.shape)

        coeff_fit_flat = fit_coefficient.flatten()
        for m in range(lmax):
            for el in range(m, lmax):
                k = cls.index_legendre(el, m)
                if m == 0:
                    v = v + coeff_fit_flat[k] * plm_fit[:, 0, el]
                else:
                    v = v + coeff_fit_flat[k] * np.cos(m * phi) \
                          * plm_fit[:, m, el] + coeff_fit_flat[k+1] \
                          * np.sin(m * phi) * plm_fit[:, m, el]

        pot_arr = np.zeros((num_lons, num_lats))
        pot_arr = np.reshape(v, pot_arr.shape) / 1000.0

        # TODO: Account for lon_shift
        # TODO: Code for lat shift! (both rarely non-0 though)
        # grid_arr[1,:] = (grid_arr[1,:] + lon_shift)

        mlat_center = grid_arr[0, :].reshape((num_lons, num_lats))
        # Set everything below the latmin as 0
        ind = np.where(abs(mlat_center) < abs(lat_min))
        pot_arr[ind] = 0

        mlon_center = grid_arr[1, :].reshape((num_lons, num_lats))
        # Invert for Southern maps
        mlat_center = mlat_center * hemisphere.value

        return mlat_center, mlon_center, pot_arr

    @classmethod
    def calculate_potentials_pos(cls, mlat, mlon, fit_coefficient: list,
                                 lat_min: list, lat_shift: int = 0,
                                 lon_shift: int = 0, fit_order: int = 6,
                                 hemisphere: Enum = Hemisphere.North,
                                 **kwargs):
        '''
        Calculates potential for a specific magnetic latitude and longitude,
        or list of mlats and mlons

        Parameters
        ----------
            mlat: float or List[float]
                Magnetic latitudes of the positions of interest
            mlon: float or List[float]
                Magnetic latitudes of the positions of interest
            fit_coefficient: List[float]
                Value of the coefficient
            lat_min: List[float]
                Minimum latitude that will be evaluated
                Not to be confused with 'lowlat'
            lat_shift: int
                Generic shift in latitude from map file
                default: 0
            lon_shift: int
                Generic shift in longitude from map file
                default: 0
            fit_order: int
                order of the fit
                default: 6
            hemisphere: Enum
                Describes the hemisphere, North or South
                default: Hemisphere.North

        Returns
        -------
            v: List[float]
                list of potentials at given position(s) in kV
        '''
        # Check input is in correct format and lengths
        if not isinstance(mlat, list):
            mlat = [mlat]
        if not isinstance(mlon, list):
            mlon = [mlon]
        if not len(mlat) == len(mlon):
            raise ValueError('mlat and mlon must be the same length.')

        # Lowest latitude to calculate potential to
        theta_max = np.radians(90 - np.abs(lat_min) + 10) * hemisphere.value

        # Convert grid vals to spherical coords
        theta = np.radians(90.0 - np.abs(mlat))
        phi = np.radians(mlon)

        # Adjusted/Normalised values (runs 0 - pi)
        alpha = np.pi / theta_max
        x = np.cos(alpha * theta)

        # Legendre Polys
        for j, xj in enumerate(x):
            plm_tmp = special.lpmn(fit_order, fit_order, xj)
            if j == 0:
                plm_fit = np.append([plm_tmp[0]], [plm_tmp[0]], axis=0)
            else:
                plm_fit = np.append(plm_fit, [plm_tmp[0]], axis=0)
        # Remove first element as it is duplicated to start off the array
        plm_fit = np.delete(plm_fit, 0, 0)

        # Eval the potential
        lmax = plm_fit.shape
        lmax = lmax[1]
        v = np.zeros(phi.shape)

        coeff_fit_flat = fit_coefficient.flatten()
        for m in range(lmax):
            for el in range(m, lmax):
                k = cls.index_legendre(el, m)
                if m == 0:
                    v = v + coeff_fit_flat[k] * plm_fit[:, 0, el]
                else:
                    v = v + coeff_fit_flat[k] * np.cos(m * phi) \
                        * plm_fit[:, m, el] + coeff_fit_flat[k + 1] \
                        * np.sin(m * phi) * plm_fit[:, m, el]

        # Convert from V to kV
        v /= 1000

        return v

    @classmethod
    def plot_potential_contours(cls, fit_coefficient: list, lat_min: list,
                                date: object, ax: object, lat_shift: int = 0,
                                lon_shift: int = 0, fit_order: int = 6,
                                hemisphere: Enum = Hemisphere.North,
                                contour_levels: list = [],
                                contour_spacing: int = None,
                                contour_color: str = 'dimgrey',
                                contour_linewidths: float = 0.8,
                                contour_fill: bool = False,
                                contour_colorbar: bool = True,
                                contour_fill_cmap: str = 'RdBu',
                                contour_colorbar_label: str = 'Potential (kV)',
                                pot_minmax_color: str = 'k',
                                pot_zmin: int = -50,
                                pot_zmax: int = 50,
                                **kwargs):
        # TODO: No evaluation of coordinate system made! May need if in
        # plotting to plot in radians/geo ect.
        '''
        Takes the grid of potentials, plots a contour plot and min and max
        potential positions

        Parameters
        ----------
            fit_coefficient: List[float]
                Value of the coefficient
            lat_min: List[float]
                Minimum latitude that will be evaluated
                Not to be confused with 'lowlat'
            date: datetime object
                Date from record
            ax: object
                matplotlib axis object
            lat_shift: int
                Generic shift in latitude from map file
                default: 0
            lon_shift: int
                Generic shift in longitude from map file
                default: 0
            fit_order: int
                order of the fit
                default: 6
            contour_levels: np.arr
                Array of values at which the contours
                are plotted
                Default: []
                Default list is defined in function due
                to length of the list, values higher or
                lower than the minimum and maximum values
                given are colored in as min and max color
                values if contour_fill=True
            contour_spacing: int
                If levels are not set explicitly, then use this value to
                set the spacing between the contour levels.
                Default is None which defaults to the max value/20
                which works out to be 5 with other default values
            contour_color: str
                Colour of the contour lines plotted
                Default: dimgrey
            contour_label: bool - NOT CURRENTLY IMPLEMENTED
                If contour_fill is True, contour labels will
                be plotted on the contour lines
                Default: True
            contour_linewidths: float
                Thickness of contour lines
                Default: 0.8
            contour_fill: bool
                Option to use filled contours rather than
                an outline. If True, contour_color and
                contour_linewidths are ignored
                If False
                Default: False
            contour_colorbar: bool
                Option to show the colorbar for the contours
                if contour_fill = True
                Default: True
            contour_fill_cmap: matplotlib.cm
                Colormap used to fill the contours if
                contour_fill is True
                Default: 'RdBu'
            contour_colorbar_label: str
                Label for the colorbar describing the
                contours if contour_fill is True
                Default: empty string ''
            pot_minmax_color: str
                Colour of the cross and plus symbols for
                minimum and maximum potentials
                Default: 'k' - black
            pot_zmin: float
                Minumum value of color map used for potential contours.
                If None, defaults to -abs(pot_arr).max()
            pot_zmax: float
                Maximum value of color map used for potential contours.
                If None, defaults to abs(pot_arr).max()
            **kwargs
                including lowlat and hemisphere for calculating
                potentials
        '''
        mlat, mlon_u, pot_arr = cls.calculate_potentials(
                             fit_coefficient, lat_min,
                             lat_shift=lat_shift, lon_shift=lon_shift,
                             fit_order=fit_order, hemisphere=hemisphere,
                             **kwargs)

        # Shift mlon to MLT
        shifted_mlts = mlon_u[0, 0] - \
            (aacgmv2.convert_mlt(mlon_u[0, 0], date) * 15)
        shifted_lons = mlon_u - shifted_mlts
        mlon = shifted_lons

        # Contained in function as too long to go into the function call
        if contour_levels == []:
            if contour_spacing is None:
                # If not set this sets the maximum number of contours to be 40
                contour_spacing = int(np.floor(np.max([abs(pot_zmin),
                                                       abs(pot_zmax)]) / 10))

            # Making the levels required, but skipping 0 as default to avoid a
            # contour at 0 position (looks weird)
            contour_levels = [*range(pot_zmin, 0, contour_spacing),
                              *range(contour_spacing,
                                     pot_zmax + contour_spacing,
                                     contour_spacing)]
        else:
            pot_zmax = np.max(contour_levels)
            pot_zmin = np.min(contour_levels)

        if contour_fill:
            # Filled contours
            norm = colors.Normalize
            norm = norm(pot_zmin, pot_zmax)
            cs = plt.contourf(np.radians(mlon), mlat, pot_arr, 2,
                              norm=norm, vmax=pot_zmax, vmin=pot_zmin,
                              levels=np.array(contour_levels),
                              cmap=contour_fill_cmap, alpha=0.5,
                              extend='both', zorder=3.0)
            if contour_colorbar is True:
                mappable = cm.ScalarMappable(norm=norm, cmap=contour_fill_cmap)
                locator = ticker.MaxNLocator(symmetric=True, min_n_ticks=3,
                                             integer=True, nbins='auto')
                ticks = locator.tick_values(vmin=pot_zmin,
                                            vmax=pot_zmax)
                cb_contour = plt.colorbar(mappable, ax=ax, extend='both',
                                          ticks=ticks, alpha=0.5)
                if contour_colorbar_label != '':
                    cb_contour.set_label(contour_colorbar_label)
            else:
                cb_contour = None
        else:
            # Contour lines only
            cs = plt.contour(np.radians(mlon), mlat, pot_arr, 2,
                             vmax=pot_zmax, vmin=pot_zmin,
                             levels=np.array(contour_levels),
                             colors=contour_color, alpha=0.8,
                             linewidths=contour_linewidths, zorder=3.0)
            cb_contour = None
            # TODO: Add in contour labels
            # if contour_label:
            #    plt.clabel(cs, cs.levels, inline=True, fmt='%d', fontsize=5)

        # Get max value of potential
        ind_max = np.where(pot_arr == pot_arr.max())
        ind_min = np.where(pot_arr == pot_arr.min())
        max_mlon = mlon[ind_max]
        max_mlat = mlat[ind_max]
        min_mlon = mlon[ind_min]
        min_mlat = mlat[ind_min]

        plt.scatter(np.radians(max_mlon), max_mlat, marker='+', s=70,
                    color=pot_minmax_color, zorder=5.0)
        plt.scatter(np.radians(min_mlon), min_mlat, marker='_', s=70,
                    color=pot_minmax_color, zorder=5.0)
        return mlat, mlon_u, pot_arr, cs, cb_contour

    @classmethod
    def find_map_record(cls, dmap_data: List[dict], start_time: dt.datetime):
        """
        looks through the data from a given map file and
        returns the record number nearest the
        passed datetime object

        Parameters
        -----------
        dmap_data: List[dict]
            the data to look through
        start_time: datetime
            the time to find the nearest record number to

        Returns
        -------
        Returns the closet record to the passed time
        """
        # recursively identify the index of the list
        # where the time is closest to the passed time
        def find_nearest_time(dmap_data: List[dict], start_time: dt.datetime,
                              start_index: int, end_index: int):
            # base case
            if start_index == end_index:
                return start_index
            # recursive case
            else:
                # find the middle index
                mid_index = int((start_index + end_index)/2)
                # if the time at the middle index is
                # greater than the passed time
                if time2datetime(dmap_data[mid_index]) > start_time:
                    # search the lower half of the list
                    return find_nearest_time(dmap_data, start_time,
                                             start_index, mid_index)
                # if the time at the middle index
                # is less than the passed time
                elif time2datetime(dmap_data[mid_index]) < start_time:
                    # search the upper half of the list
                    return find_nearest_time(dmap_data, start_time,
                                             mid_index + 1, end_index)
                # if the time at the middle index is equal to the passed time
                else:
                    # return the middle index
                    return mid_index

        return find_nearest_time(dmap_data,
                                 start_time, 0, len(dmap_data) - 1)

    @classmethod
    def plot_time_series(cls, dmap_data: List[dict],
                         parameter: Enum = TimeSeriesParams.NUM_VECTORS,
                         start_record: int = 0, end_record: int = 1,
                         start_time: dt.datetime = None,
                         end_time: dt.datetime = None,
                         potential_position: list = [],
                         ax=None, **kwargs):
        '''
        Plot time series of various map parameters

        Params
        ----------
        dmap_data: List[dict]
            List of dictionaries containing the data to be plotted
        start_record: int
            time index that we will plot from
        end_record: int
            time index that we will plot to (-1 will plot the entire period)
        start_time: dt.datetime
            Start time of the data
        end_time: dt.datetime
            End time of the data
        potential_position: list [mlon, mlat]
            Position at which the potential is to be calculated and plotted
        ax:
            matplotlib axis
        kwargs:
            keyword arguments to be passed to the plotting function
            '''
        # if no time objects are passed & no start/end record is passed
        # then plot all availaible data
        record_absent = start_time is None and end_time is None
        if start_record == 0 and end_record == 1 and record_absent:

            start_record = 0
            end_record = len(dmap_data)-1
        # determine the start and end record
        if start_time is not None and end_time is not None:
            start_record = cls.find_map_record(dmap_data, start_time)
            end_record = cls.find_map_record(dmap_data, end_time)
        else:
            start_time = time2datetime(dmap_data[start_record])
            end_time = time2datetime(dmap_data[end_record])
        # based on the parameter, plot the data
        if parameter == TimeSeriesParams.NUM_VECTORS:
            datalist = []
            timelist = []
            for records in range(start_record, end_record):
                # append the dimension of numv for each record
                # we can just uexse any of the keys with dimensionality numv
                # Try loop for partial records
                try:
                    datalist.append(len(dmap_data[records]['vector.mlat']))
                except KeyError:
                    datalist.append(np.nan)
                # now get the associated time data point per record
                timelist.append(time2datetime(dmap_data[records]))
            plt.plot(timelist, datalist, **kwargs)
            plt.ylabel('Number of Vectors')
            plt.xlabel('Time (UTC)')
            plt.title("Number of Vectors for " + str(start_time)
                      + " to " + str(end_time))
        elif parameter == TimeSeriesParams.POT:
            # This requires a position input in [mlat, mlon]
            if len(potential_position) != 2:
                warnings.warn("A valid magnetic position with format "
                              "[mlon, mlat] "
                              "is required to plot the potential "
                              "at a given location. E.G. "
                              "potential_position = [-110, 78]")
            elif (potential_position[0] <= -180 or
                  potential_position[0] >= 180 or
                  potential_position[1] <= -90 or
                  potential_position[1] >= 90):
                warnings.warn("A valid magnetic position with format "
                              "[mlon, mlat] "
                              "is required to plot the potential "
                              "at a given location. E.G. "
                              "potential_position = [-110, 78]")
            else:
                datalist = []
                timelist = []
                for records in range(start_record, end_record):
                    # Calculate potential
                    pot = cls.calculate_potentials_pos(
                            potential_position[1],
                            potential_position[0],
                            dmap_data[records]['N+2'],
                            dmap_data[records]['latmin'],
                            dmap_data[records]['lat.shft'],
                            dmap_data[records]['lon.shft'],
                            dmap_data[records]['fit.order'],
                            Hemisphere(dmap_data[records]['hemisphere']))
                    datalist.append(pot)
                    timelist.append(time2datetime(dmap_data[records]))
                plt.plot(timelist, datalist, **kwargs)
                plt.ylabel('Potential (kV)')
                plt.xlabel('Time (UTC)')
                plt.title('Potential at ' + str(potential_position) +
                          ' for ' + str(start_time) + " to " +
                          str(end_time))
        elif parameter is not None:
            datalist = []
            timelist = []
            for records in range(start_record, end_record):
                datalist.append(dmap_data[records][parameter.value])
                timelist.append(time2datetime(dmap_data[records]))
            plt.plot(timelist, datalist, **kwargs)
            plt.ylabel(parameter.value)
            plt.xlabel('Time (UTC)')
            plt.title(parameter.value + ' for ' + str(start_time)
                      + " to " + str(end_time))



# (C) Copyright SuperDARN Canada, University of Saskatchewan
# Author: Marina Schmidt
#
# Disclaimer:
# pyDARN is under the LGPL v3 license found in the root directory LICENSE.md
# Everyone is permitted to copy and distribute verbatim copies of this license
# document, but changing it is not allowed.
#
# This version of the GNU Lesser General Public License incorporates the terms
# and conditions of version 3 of the GNU General Public License,
# supplemented by the additional permissions listed below.
#
# Modifications:
#

# Standard accepted IUGG global average of Earth's radius (km)
Re = 6371
# Earth radius at the equator (km)
EARTH_EQUATORIAL_RADIUS = 6378.137

# Speed of light (m/s)
C = 2.99792458e+8



# Copyright (C) 2021 SuperDARN Canada, University of Saskatchewan
# Author: Daniel Billett
#
# Modifications:
# 2022-03-11 MTS added enums for projection function calls
# 2022-03-22 MTS removed coastline call and added grid lines to cartopy
#                plotting
# 2022-05-20 CJM added options to plot coastlines
# 2022-06-13 Elliott Day don't create new ax if ax passed in to Projs
# 2023-02-22 CJM added options for nightshade
# 2023-08-11 RR added crude check for unmodified axes, handle both hemispheres
# 2024-05-15 CJM refactored geographic axes to add plot zoom and center, 
#            and added the geomagnetic version to do the same
# 2024-07-10 CJM removed cartopy logic to allow full dependency
#
# Disclaimer:
# pyDARN is under the LGPL v3 license found in the root directory LICENSE.md
# Everyone is permitted to copy and distribute verbatim copies of this license
# document, but changing it is not allowed.
#
# This version of the GNU Lesser General Public License incorporates the terms
# and conditions of version 3 of the GNU General Public License,
# supplemented by the additional permissions listed below.
"""
Code which generates axis objects for use in plotting functions
"""
import aacgmv2
import cartopy.crs as ccrs
import cartopy.feature as cfeature
from cartopy.feature.nightshade import Nightshade
import enum
import matplotlib.pyplot as plt
from matplotlib import axes
import matplotlib.ticker as mticker
import numpy as np

from pydarn import (Hemisphere, Re, nightshade_warning)


def convert_geo_coastline_to_mag(geom, date, alt: float = 0.0, mag_lon: bool = False):
    """
    Takes the geometry object of coastlines and converts
    the coordinates into AACGM_MLT
    Only required usage is for cartopys NaturalEarthFeature
    at 110m resolution only
    Parameters
    ----------
    geom: Shapely Geometry object
        A list/collection of geometry objects
    date: datetime object
        Date of required plot
    alt: float
        Altitude in km
        Default 0 (sea level) for coastlines
    mag_lon: bool
        Set true to return magnetic longitude, not MLT
    """
    [mlats, lon_mag, _] = \
        aacgmv2.convert_latlon_arr(geom.coords.xy[1], geom.coords.xy[0], alt,
                                       date, method_code='G2A')

    # Finds the first not nan value to calculate the mlt shift
    # Substitutes NaN if not found which results in no data to plot
    # aacgmv2 will return NaNs as there are some lat/lon combinations that
    # do not correspond to a geomagnetic position.
    notnan_lon = next((x for x in lon_mag if x == x), float('NaN'))

    # Shift to MLT
    shifted_mlts = notnan_lon - \
                    (aacgmv2.convert_mlt(notnan_lon, date) * 15)
    shifted_lons = lon_mag - shifted_mlts

    if mag_lon:
        mlons = shifted_lons
    else:
        mlons = np.radians(shifted_lons)
    # Return geometry object
    return type(geom)(list(zip(mlons, mlats)))


def axis_geomagnetic(date, ax: axes.Axes = None, lowlat: int = 30,
                    hemisphere: Hemisphere = Hemisphere.North,
                    coastline: bool = False, cartopy_scale: str = '110m',
                    coastline_color: str = 'k',
                    coastline_linewidth: float = 0.5,
                    nightshade: int = 0, grid_lines: bool = True,
                    plot_center: list = None,
                    plot_extent: list = None, **kwargs):
    """
    Sets up the cartopy orthographic plot axis object, for use in
    various other functions. This plot assumes you are giving geoMAGNETIC
    values for plotting, so extra additional cartopy functions such as
    coastlines do not work in this context, you must convert to geomagnetic.
    
    Parameters
    ----------
        date: datetime object
            Date of required plot
        ax: matplotlib.axes.Axes
            Pre-defined axis object to pass in, must be
            geological projection
            Default: Generates a geographical projection for the user
            with geographic latitude/longitude labels
        lowlat: int
            Lower geographic latitude boundary for the geographic plot
            Default: 30
        hemisphere: enum
            Hemisphere of geographic projection. Can be Hemisphere.North or
            Hemisphere.South for northern and southern hemispheres,
            respectively
            Default: Hemisphere.North
        grid_lines: bool
            add latitude/longtitude lines with labels to the plot
            Default: True
        coastline: bool
            Set to true to overlay coastlines with cartopy
        nightshade: int
            Altitude above surface for calculating regions shadowed from Sun.
        plot_center: list [float, float]
            Longitude and Latitude of the desired center of the plot
            Plot will still plot even if data is on the other side of the Earth
            Remember to include negative latitude for Southern Hemisphere
            Setting this option will change the rotation and 12 MLT may no
            longer be at the bottom of the plot
            Default: None
            Example: [-90, 60] will show the Earth centered on Canada
        plot_extent: list [float, float]
            Plotting extent in terms of a percentage of Earth shown in
            the x and y plotting field 
            Default: None
            Example: [30, 50] shows a plot centered on the pole or specified
                     plot_center coord that shows 30% of the Earth in x and 
                     50% of the Earth in y. See tutorials for plotted example.
    """
    if plot_center is None:
        # If no center for plotting is given, default to pole
        if hemisphere == Hemisphere.North:
            pole_lat = 90
            # Keep 0 MLT at the bottom of plot by default
            lon = 0
            lat = abs(lowlat)
        else:
            pole_lat = -90
            # Keep 0 MLT at the bottom of plot by default
            lon = 180
            lat = -abs(lowlat)
        if ax is None:
            proj = ccrs.Orthographic(lon, pole_lat)
            ax = plt.axes(projection=proj)
            if plot_extent is not None and len(plot_extent) == 2:
                # Padding in % of Earth radius
                padx = (plot_extent[0] / 100) * Re*1000
                pady = (plot_extent[1] / 100) * Re*1000
                ax.set_extent(extents=(-padx, padx, -pady, pady), crs=proj)
            else:
                extent = abs(proj.transform_point(lon, lat, ccrs.PlateCarree())[1])
                ax.set_extent(extents=(-extent, extent, -extent, extent), crs=proj)

    else:
        # If the center of the plot is given- shift it around
        lon = plot_center[0]
        lat = plot_center[1]
        if ax is None:
            proj = ccrs.Orthographic(lon, lat)
            ax = plt.axes(projection=proj)
            if plot_extent is not None and len(plot_extent) == 2:
                # Padding in % of Earth radius
                padx = (plot_extent[0] / 100) * Re*1000
                pady = (plot_extent[1] / 100) * Re*1000
                ax.set_extent(extents=(-padx, padx, -pady, pady), crs=proj)
            else:
                # If size is not set or set incorrectly, show  whole
                # hemisphere centered on plot_center
                ax.set_global()

    def mlt_ticklabels(x, pos):
        ''' function for converting mlon to MLT for sole use in tick labels
            for a FuncFormatter operation
        '''
        # Calculate MLT from the mlon values (already shifted)
        MLT = aacgmv2.convert_mlt(x, date)
        # Undo the original shift done to the MLT values to get back to 0-23
        unshiftedMLT = MLT + shift
        # Make sure it is cyclical from 0-24
        if unshiftedMLT < 0:
            unshiftedMLT = unshiftedMLT + 24
        elif unshiftedMLT >= 24:
            unshiftedMLT = unshiftedMLT - 24
        return str(int(np.round(unshiftedMLT))) + ' MLT'

    if grid_lines:
        # Mag Lon gridlines
        gl = ax.gridlines(draw_labels=True)
        # Positions of gridl ines required in MLT
        MLT = [0, 3, 6, 9, 12, 15, 18, 21]
        # Shift MLT to plotting position
        MLT_gridlines = MLT - (MLT[0] - aacgmv2.convert_mlt(MLT[0], date))
        shift = MLT[0] - aacgmv2.convert_mlt(MLT[0], date)
        # Convert shifted MLT to mlon
        mlon_gridlines = aacgmv2.convert_mlt(MLT_gridlines, date, m2a=True)
        # Put grid lines on the MLT positions
        gl.xlocator = mticker.FixedLocator(mlon_gridlines)
        # format the labels to show MLT values
        gl.xformatter = plt.FuncFormatter(mlt_ticklabels)

    if coastline:
        # Read in the geometry object of the coastlines
        cc = cfeature.NaturalEarthFeature('physical', 'coastline',
                                          cartopy_scale, color='k',
                                          zorder=2.0)
        # Convert geometry object coordinates to MLT
        geom_mag = []
        for geom in cc.geometries():
            if geom.__class__.__name__ == 'MultiLineString':
                for g in geom.geoms:
                    geom_mag.append(convert_geo_coastline_to_mag(g, date, mag_lon=True))
            else:
                geom_mag.append(convert_geo_coastline_to_mag(geom, date, mag_lon=True))
        cc_mag = cfeature.ShapelyFeature(geom_mag, ccrs.Geodetic(),
                                         color='k', zorder=2.0)

        # Plot each geometry object
        for geom in cc_mag.geometries():
            plt.plot(*geom.coords.xy, color=coastline_color,
                     linewidth=coastline_linewidth, zorder=2.0,
                     transform = ccrs.Geodetic())

    if nightshade:
        nightshade_warning()

    return ax, ccrs


def axis_geomagnetic_polar(date, ax: axes.Axes = None, lowlat: int = 30,
                    hemisphere: Hemisphere = Hemisphere.North,
                    coastline: bool = False, cartopy_scale: str = '110m',
                    coastline_color: str = 'k',
                    coastline_linewidth: float = 0.5,
                    nightshade: int = 0, **kwargs):
    """
    Sets up the polar plot matplotlib axis object, for use in
    various other functions. Magnetic latitude - magnetic local
    time projection.

    This projection is defunct now MAG exists, however is kept available for
    compatibility and non-cartopy options. 

    Parameters
    -----------
        date: datetime object
            Date of required plot. Only required if using coastlines
            (for AACGM conversion)
        ax: matplotlib.axes.Axes
            Pre-defined axis object to pass in, must be
            polar projection
            Default: Generates a polar projection for the user
            with MLT/latitude labels
        lowlat: int
            Lower AACGM latitude boundary for the polar plot
            Default: 30
        hemisphere: enum
            Hemisphere of polar projection. Can be Hemisphere.North or
            Hemisphere. South for northern and southern hemispheres,
            respectively
            Default: Hemisphere.North
        coastline: bool
            Set to true to overlay coastlines with cartopy. Requires
            date.
        nightshade: int
            Altitude above surface for calculating regions shadowed from Sun.
            Not supported for this projection.
        cartopy_scale: str
            string corresponding with the scale to plot the coastlines at
            options: '110m', '50m', '10m'
        coastline_color: str
            color of the coastline outline
        coastline_linewidth: float
            linewidth of the coastline feature
            default 0.5
    """

    if ax is None:
        ax = plt.axes(polar=True)

        # Set upper and lower latitude limits (pole and lowlat)
        if hemisphere == Hemisphere.North:
            ax.set_ylim(90, lowlat)
            ax.set_yticks(np.arange(lowlat, 90, 10))
        else:
            # If hemisphere is South, lowlat must be negative
            ax.set_ylim(-90, -abs(lowlat))
            ax.set_yticks(np.arange(-abs(lowlat), -90, -10))

        # Locations of tick marks. Will be customisable in future
        ax.set_xticks([0, np.radians(45), np.radians(90), np.radians(135),
                       np.radians(180), np.radians(225), np.radians(270),
                       np.radians(315)])

        # Tick labels will depend on coordinate system
        ax.set_xticklabels(['00', '', '06', '', '12', '', '18', ''])
        ax.set_theta_zero_location("S")
    else:
        if ax.get_ylim() == (0.0, 1.0):
            # Set upper and lower latitude limits (pole and lowlat)
            if hemisphere == Hemisphere.North:
                ax.set_ylim(90, lowlat)
                ax.set_yticks(np.arange(lowlat, 90, 10))
            else:
                # If hemisphere is South, lowlat must be negative
                ax.set_ylim(-90, -abs(lowlat))
                ax.set_yticks(np.arange(-abs(lowlat), -90, -10))

    if coastline:
        # Read in the geometry object of the coastlines
        cc = cfeature.NaturalEarthFeature('physical', 'coastline',
                                          cartopy_scale, color='k',
                                          zorder=2.0)
        # Convert geometry object coordinates to MLT
        geom_mag = []
        for geom in cc.geometries():
            if geom.__class__.__name__ == 'MultiLineString':
                for g in geom.geoms:
                    geom_mag.append(convert_geo_coastline_to_mag(g, date))
            else:
                geom_mag.append(convert_geo_coastline_to_mag(geom, date))
        cc_mag = cfeature.ShapelyFeature(geom_mag, ccrs.PlateCarree(),
                                         color='k', zorder=2.0)
        # Plot each geometry object
        for geom in cc_mag.geometries():
            plt.plot(*geom.coords.xy, color=coastline_color,
                     linewidth=coastline_linewidth, zorder=2.0)

    if nightshade:
        nightshade_warning()

    return ax, None


def axis_geographic(date, ax: axes.Axes = None,
                    hemisphere: Hemisphere = Hemisphere.North,
                    lowlat: int = 30, grid_lines: bool = True,
                    coastline: bool = False, nightshade: int = 0,
                    cartopy_scale: str = '110m', coastline_color: str = 'k',
                    coastline_linewidth: float = 0.5,
                    plot_center: list = None,
                    plot_extent: list = None, **kwargs):

    """

    Sets up the cartopy orthographic plot axis object, for use in
    various other functions. Geographic projection.

    Parameters
    -----------
        date: datetime object
            Date of required plot
        ax: matplotlib.axes.Axes
            Pre-defined axis object to pass in, must be
            geological projection
            Default: Generates a geographical projection for the user
            with geographic latitude/longitude labels
        lowlat: int
            Lower geographic latitude boundary for the geographic plot
            Default: 30
        hemisphere: enum
            Hemisphere of geographic projection. Can be Hemisphere.North or
            Hemisphere.South for northern and southern hemispheres,
            respectively
            Default: Hemisphere.North
        grid_lines: bool
            add latitude/longtidue lines with labels to the plot
            Default: True
        coastline: bool
            Set to true to overlay coastlines with cartopy
        nightshade: int
            Altitude above surface for calculating regions shadowed from Sun.

        cartopy_scale: str
            string corresponding with the scale to plot the coastlines at
            options: '110m', '50m', '10m'
        coastline_color: str
            color of the coastline outline
        coastline_linewidth: float
            linewidth of the coastline feature
            default 0.5
        plot_center: list [float, float]
            Longitude and Latitude of the desired center of the plot
            Plot will still plot if data is on the other side of the Earth
            Remember to include negative latitude for Southern Hemisphere
            Default: None
            Example: [-90, 60] will show the Earth centered on Canada
        plot_extent: list [float, float]
            Plotting extent in terms of a percentage of Earth shown in
            the x and y plotting field 
            Default: None
            Example: [30, 50] shows a plot centered on the pole or specified
                     plot_center coord that shows 30% of the Earth in x and 
                     50% of the Earth in y. See tutorials for plotted example.
    """
    if plot_center is None:
        # If no center for plotting is given, default to pole and rotate for 0
        # degrees longitude
        if hemisphere == Hemisphere.North:
            pole_lat = 90
            lon = 0
            lat = abs(lowlat)
        else:
            pole_lat = -90
            lon = 0
            lat = -abs(lowlat)
        if ax is None:
            proj = ccrs.Orthographic(lon, pole_lat)
            ax = plt.axes(projection=proj)
            if plot_extent is not None and len(plot_extent) == 2:
                # Padding in % of Earth radius
                padx = (plot_extent[0] / 100) * Re*1000
                pady = (plot_extent[1] / 100) * Re*1000
                ax.set_extent(extents=(-padx, padx, -pady, pady), crs=proj)
            else:
                extent = abs(proj.transform_point(lon, lat, ccrs.PlateCarree())[1])
                ax.set_extent(extents=(-extent, extent, -extent, extent), crs=proj)
    else:
        # If the center of the plot is given- shift it around
        lon = plot_center[0]
        lat = plot_center[1]
        if ax is None:
            proj = ccrs.Orthographic(lon, lat)
            ax = plt.axes(projection=proj)
            if plot_extent is not None and len(plot_extent) == 2:
                # Padding in % of Earth radius
                padx = (plot_extent[0] / 100) * Re*1000
                pady = (plot_extent[1] / 100) * Re*1000
                ax.set_extent(extents=(-padx, padx, -pady, pady), crs=proj)
            else:
                # If size is not set or set incorrectly, show  whole
                # hemisphere centered on plot_center
                ax.set_global()

    if grid_lines:
        ax.gridlines(draw_labels=True)

    if coastline:
        ax.coastlines(resolution=cartopy_scale, color=coastline_color,
                      linewidth=coastline_linewidth)

    if nightshade:
        refraction_value = -np.degrees(np.arccos(Re / (Re + nightshade)))
        ns = Nightshade(date, refraction=refraction_value, alpha=0.1)
        ax.add_feature(ns)

    return ax, ccrs


class Projs(enum.Enum):
    """
    class of projections that pydarn can do

    Enumerators
    ------------
    POLAR: axes_polar
    GEO: axes_geological
    """
    POLAR = (axis_geomagnetic_polar, )
    GEO = (axis_geographic, )
    MAG = (axis_geomagnetic, )

    # Need this to make the functions callable
    def __call__(self, *args, **kwargs):
        return self.value[0](*args, **kwargs)



# (C) Copyright SuperDARN Canada, University of Saskatchewan
# Author: Carley Martin
#
# Disclaimer:
# pyDARN is under the LGPL v3 license found in the root directory LICENSE.md
# Everyone is permitted to copy and distribute verbatim copies of this license
# document, but changing it is not allowed.
#
# This version of the GNU Lesser General Public License incorporates the terms
# and conditions of version 3 of the GNU General Public License,
# supplemented by the additional permissions listed below.
#
# Modifications:
#


class Citations():
    '''
    A class to contain the different citations for pyDARN use
    
    Parameters
    ----------
    citations with format authorYYYY = str
    '''
    # Groundscatter equation
    bristow1994 = 'Bristow, W. A. et al, 1994, 10.1029/93JA01470 \n'
    # AACGMv3 Wrapper:
    burrell2023 = 'Burrell, A. G. et al, 2023, 10.5281/zenodo.7621545 \n'
    # Cartopy:
    cartopy2015 = 'Met Office, 2010-2015, scitools.org.uk/cartopy \n'
    # Chisham model/ Standard model:
    chisham2008 = 'Chisham, G. 2008, 10.5194/angeo-26-823-2008 \n'
    #pydarn citation:
    pydarn2023 = 'DVWG, 2023, 10.5281/zenodo.3727269 \n'
    #pydarnio citation:
    pydarnio2023 = 'DVWG, 2023, 10.5281/zenodo.4009470 \n'
    # AACGMv2 Paper:
    shepherd2014 = 'Shepherd, S. G., 2014, 10.1002/2014JA020264 \n'
    # elevation angle claculations:
    shepherd2017 = 'Shepherd, S. G., 2017, 10.1002/2017RS006348 \n'
    # GSMR Options:
    thomas2022 = 'Thomas, E. G. et al, 2022, 10.1029/2022RS007429 \n'


    @classmethod
    def print_citations(self):
        print("\nIf using pyDARN produced plots in publications please be "
              "aware of the following citations that may have been used to"
              " produce your plot and should be included in your publication:"
              "\n\n",
              "Ground Scatter Mapped Range: ", self.bristow1994,
              "AACGMv2 Wrapper:             ", self.burrell2023,
              "Chisham Virtual Height Model:", self.chisham2008,
              "pyDARN Software:             ", self.pydarn2023,
              "pyDARNio Software:           ", self.pydarnio2023,
              "Cartopy:                     ", self.cartopy2015,
              "AACGMv2 article:             ", self.shepherd2014,
              "Elevation Angle Calculations:", self.shepherd2017,
              "Ground Scatter Mapped Range: ", self.thomas2022,
              "\nFurther information can be found at"
              " https://pydarn.readthedocs.io/en/main/user/citing/\n")


    @classmethod
    def print_acknowledgements(self):
        print("\nAcknowledgement required for SuperDARN data use:\n\n"
              "\x1B[3m"
              "The authors acknowledge the use of SuperDARN data. "
              "SuperDARN is a collection of radars funded by "
              "national scientific funding agencies of Australia, "
              "Canada, China, France, Italy, Japan, Norway, South "
              "Africa, United Kingdom and the United States of "
              "America."
              "\x1B[0m \n\n"
              "During your study, if using data from individual "
              "radars only, please contact the Principal Investigator "
              "(PI) of that radar about potential co-authorship or "
              "appropriate acknowledgments."
              "\n")



# Copyright (C) 2020 SuperDARN Canada, University of Saskatchewan
# Author: Marina Schmidt
#
# Modifications:
#
# Disclaimer:
# pyDARN is under the LGPL v3 license found in the root directory LICENSE.md
# Everyone is permitted to copy and distribute verbatim copies of this license
# document, but changing it is not allowed.
#
# This version of the GNU Lesser General Public License incorporates the terms
# and conditions of version 3 of the GNU General Public License,
# supplemented by the additional permissions listed below.

import datetime as dt
import matplotlib.pyplot as plt
import pytest
import warnings

import pydarn


data = pydarn.SuperDARNRead('test/data/test.grd').read_grid()


class TestGrid_defaults:

    def test_grid_defaults(self):
        """ """
        with warnings.catch_warnings(record=True):
            pydarn.Grid.plot_grid(data)

@pytest.mark.parametrize('colorbar', [False])
@pytest.mark.parametrize('colorbar_label', 'green')
@pytest.mark.parametrize('title', [False])
@pytest.mark.parametrize('cmap', [plt.get_cmap('rainbow')])
@pytest.mark.parametrize('record', [2])
@pytest.mark.parametrize('ranges', [(5,70)])
@pytest.mark.parametrize('time_delta', [2])
@pytest.mark.parametrize('parameter', ['wdt', 'pwr'])
@pytest.mark.parametrize('radar_location', [False])
@pytest.mark.parametrize('radar_label', [True])
@pytest.mark.parametrize('start_time', [dt.datetime(2021, 2, 5, 12, 5)])
@pytest.mark.parametrize('zmin', [0])
@pytest.mark.parametrize('zmax', [100])
@pytest.mark.parametrize('len_factor', [150.0, 100.0])
class TestGrid:

    def test_plot_grid(self, record, colorbar, colorbar_label,
                       title, cmap, ranges, time_delta,
                       parameter, radar_location, radar_label,
                       start_time, zmin, zmax, len_factor):
        """ this test will give bare minimum input needed for """
        with warnings.catch_warnings(record=True):
            pydarn.Grid.plot_grid(data, ranges=ranges,
                                  record=record, time_delta=time_delta,
                                  parameter=parameter,
                                  radar_location=radar_location,
                                  radar_label=radar_label,
                                  start_time=start_time, zmin=zmin,
                                  zmax=zmax, len_factor=len_factor)
        plt.close('all')



# (C) Copyright 2021 SuperDARN Canada, University of Saskatchewan
# Author(s): Marina Schmidt
#
# Disclaimer:
# pyDARN is under the LGPL v3 license found in the root directory LICENSE.md
# Everyone is permitted to copy and distribute verbatim copies of this license
# document, but changing it is not allowed.
#
# This version of the GNU Lesser General Public License incorporates the terms
# and conditions of version 3 of the GNU General Public License,
# supplemented by the additional permissions listed below.
#
# Modifications:
# 2022-08-04 CJM added HALF_SLANT option and gate2halfslant method
# 2023-09-14 CJM moved GSMR to GSMR_BRISTOW and used new GSMR alg
# 2023-12-15 RAR added TIME_OF_FLIGHT option and gate2timeofflight method

import enum
import numpy as np
import warnings

from pydarn import Re, C, standard_warning_format

warnings.formatwarning = standard_warning_format


def gate2timeofflight(rxrise: int = 0, range_gate: int = 0, frang: int = 180,
                      rsep: int = 45, nrang: int = None, center: bool = True,
                      **kwargs):
    """
    Calculate the time of flight for each range gate.

    Parameters
    ----------
    frang: int
        range from the edge of first the gate to the radar [km]
        This should be given in fitacf record of the control program
    rsep: int
        Radar separation of the gates. Determined by control program.
    rxrise: int
        Use hardware value for this, avoid data file values
    gate: int
        range gate to determine the slant range [km], if nrang
        is None
        default: 0
    nrang: int
        max number of range gates in the list of records. If
        not None, will calculate all slant ranges
        default: None
    center: boolean
        Calculate the slant range in the center of range gate
        or edge

    Returns
    -------
    tof: np.array
        returns an array of times of flight, in ms
    """
    # lag to the first range gate in microseconds
    # 2 - two times for there and back
    distance_factor = 2.0
    # C - speed of light m/s to km/ms
    speed_of_light = C * 0.001 * 1e-3
    lag_first = frang * distance_factor / speed_of_light
    # sample separation in microseconds
    sample_sep = rsep * distance_factor / speed_of_light
    # Range offset
    # If center is true, calculate at the center
    if center:
        # 0.5 offset to the centre of the range gate instead of edge
        range_offset = -0.5 * sample_sep
    else:
        range_offset = 0.0
    # Now calculate time of flight in ms
    if nrang is None:
        tof = lag_first - rxrise + range_gate * sample_sep + range_offset
    else:
        tof = np.zeros(nrang + 1)
        for gate in range(nrang + 1):
            tof[gate] = lag_first - rxrise + gate * sample_sep + range_offset
    return tof


def gate2halfslant(**kwargs):
    """
    Calculate the slant range divided by 2 for each range gate

    Parameters
    ----------
        None

    Returns
    -------
        half_slant : np.array
            returns an array of slant ranges divided by two for the radar
    """
    slant_range = gate2slant(**kwargs)
    half_slant = slant_range / 2
    return half_slant


def gate2gs_bristow(virtual_height: float = 250, **kwargs):
    """
    Calculate the ground scatter mapped range (km) for each slanted range
    for SuperDARN data. This function is based on the Ground Scatter equation
    from Bristow paper at https://doi.org/10.1029/93JA01470 on page 325
    Parameters
    ----------
        virtual_height: float
            virtual height
            default:  250

    Returns
    -------
        ground_scatter_mapped_ranges : np.array
            returns an array of ground scatter mapped ranges for the radar
    """
    slant_ranges = gate2slant(**kwargs)
    ground_scatter_mapped_ranges = Re * np.arcsin(np.sqrt((slant_ranges**2 / 4)
                                                  - (virtual_height**2)) / Re)
    # Check to see if there is an issue with the sqrt and
    # give user a warning if so, these values will be dealt with in
    # the individual plotting algs as we need to return the full array
    # of values for the complete beam*range gate array
    if any(np.isfinite(ground_scatter_mapped_ranges)):
        warnings.warn("Warning: Be aware that the range estimation"
                      " you have chosen has calculated some infinite"
                      " values. These values will not be plotted."
                      " You may use RangeEstimation.GSMR to avoid this"
                      " issue.")
    return ground_scatter_mapped_ranges


def gate2groundscatter(virtual_height: float = 250, hop: float = 0.5,
                       **kwargs):
    """
    Calculate the ground scatter mapped range (km) for each slanted range
    for SuperDARN data. This function is based on the Ground Scatter equation
    discussed in the issue github.com/SuperDARN/pydarn/issues/257
    Parameters
    ----------
        virtual_height: float
            virtual height
            default:  250
        hop: float
            hop number of returning data
            default: 0.5

    Returns
    -------
        ground_scatter_mapped_ranges : np.array
            returns an array of ground scatter mapped ranges for the radar
    """
    slant_ranges = gate2slant(**kwargs)

    num = - Re**2 - (Re + virtual_height)**2\
        + (slant_ranges/2 * (0.5 / hop))**2
    den = 2 * Re * (Re + virtual_height)
    ground_scatter_mapped_ranges = (hop/0.5) * Re * np.arccos(- num / den)

    return ground_scatter_mapped_ranges


def gate2slant(rxrise: int = 0, range_gate: int = 0, frang: int = 180,
               rsep: int = 45, nrang: int = None, center: bool = False,
               **kwargs):
    """
    Calculate the slant range (km) for each range gate for SuperDARN data

    Parameters
    ----------
        frang: int
            range from the edge of first the gate to the radar [km]
            This should be given in fitacf record of the control program
        rsep: int
            Radar separation of the gates. Determined by control program.
        rxrise: int
            Use hardware value for this, avoid data file values
        gate: int
            range gate to determine the slant range [km], if nrang
            is None
            default: 0
        nrang: int
            max number of range gates in the list of records. If
            not None, will calculate all slant ranges
            default: None
        center: boolean
            Calculate the slant range in the center of range gate
            or near-left corner see also: gate2geographic_location in
            coordinates module
            default: False (return corner values)

    Returns
    -------
        slant_ranges : np.array
            returns an array of slant ranges for the radar
    """
    # lag to the first range gate in microseconds
    # 0.3 - speed of light (km/us)
    # 2 - two times for there and back
    distance_factor = 2.0
    # C - speed of light m/s to km/us
    speed_of_light = C * 0.001 * 1e-6
    lag_first = frang * distance_factor / speed_of_light
    # sample separation in microseconds
    sample_sep = rsep * distance_factor / speed_of_light

    # Across fov direction is corrected in coordinates module already
    # 0.5 off set to the centre of the range gate instead of edge
    # This assumes that frang is to the center of the range gate
    if center:
        range_offset = 0.0
    else:
        range_offset = -0.5 * rsep

    # Now calculate slant range in km
    if nrang is None:
        slant_ranges = (lag_first - rxrise +
                        range_gate * sample_sep) * speed_of_light /\
                distance_factor + range_offset
    else:
        slant_ranges = np.zeros(nrang+1)
        for gate in range(nrang+1):
            slant_ranges[gate] = (lag_first - rxrise +
                                  gate * sample_sep) * speed_of_light /\
                    distance_factor + range_offset
    return slant_ranges


class RangeEstimation(enum.Enum):
    """
    Range_Estimation class is to list the current range gate estimations
    a user can pick from

    enumerators:
        RANGE_GATE: range gates
        SLANT_RANGE: slant range (km)
        GSMR: ground scatter mapped range (km)
    """

    RANGE_GATE = enum.auto()
    SLANT_RANGE = (gate2slant,)
    HALF_SLANT = (gate2halfslant,)
    TIME_OF_FLIGHT = (gate2timeofflight,)
    GSMR = (gate2groundscatter,)
    GSMR_BRISTOW = (gate2gs_bristow,)

    # Need this to make the functions callable
    def __call__(self, *args, **kwargs):
        return self.value[0](*args, **kwargs)
